import os
import platform
import sys
import builtins
import getpass
import glob
import re
from importlib import import_module

try:
    import traceback
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import TerminalFormatter

    use_colored_traceback = True
except:
    use_colored_traceback = False

from tklib.tkobject import tkObject
from tklib.tkparams import tkParams
from tklib.tkplugin import tkModule
from tklib.tkutils import terminate, getarg, getintarg, getfloatarg, pint, pfloat, pconv_by_type, replace_path
from tklib.tkutils import convert_by_vars, save_csv, del_quote, split_file_path

#=========================
# Application base class
#=========================


class tkApplication(tkObject):
    def __init__(self, usage_str = "", _globals = None, locals = None, 
                 use_user_inifile = False, inifile_dirs = ['ini'], create_inidir = False,
                 **args):
        super().__init__()

        self.os_name          = platform.system()
        self.args_opt_inf     = {}
        self.args_idx_inf     = []
        self.args_vars        = None

        self.set_usage(usage_str)
   
        self.phrases     = {}

        self.user            = getpass.getuser()
        self.user_nospace    = self.user.replace(' ', '')

        self.argv            = sys.argv
        self.script_path     = sys.argv[0]
        self.script_fullpath = os.path.abspath(self.script_path)
        dirname, basename, filebody, ext = split_file_path(self.script_fullpath)
        self.script_dir      = dirname
        self.script_filebody = filebody
#        self.script_dir      = os.path.splitext(self.script_fullpath)[0]

        self.ini_dir = self.get_inifile_dir(use_user_inifile = use_user_inifile,
                             inifile_dirs = inifile_dirs, create_inidir = create_inidir,
                              defval = self.script_dir)
        self.inifile = self.get_inifile_path(self.ini_dir, use_user_inifile = use_user_inifile)

        self.redirects      = []
        self.print_original = print
        
        self.globals     = globals
        self.locals      = locals
        self.modules     = []

# Configuration parameters
        self.configparams   = tkParams(app = self)
        self.config         = self.configparams
        self.config.lang    = 'en'
        self.config.inifile = self.inifile
        self.config.ini_dir = os.path.dirname(os.path.abspath(self.config.inifile))

# Application global variables
        self.params = tkParams(app = self)

# Script variables
        self.svars  = tkParams(app = self)

        self.update(**args)

    def __del__(self):
        if self.get('redirects', None):
            for target in self.redirects:
                if type(target) is str:
                    pass
                else:
                    try:
                        target.close()
                    except:
                        pass

#        super(tkObject, self).__del__()

    def __str__(self):
        return self.ClassPath()

    def exception_hook(self, type, value, tb, output = 'stderr', display_type = 'colored', pause_at_terminate = True):
        if not use_colored_traceback:
            return

        tbtext    = ''.join(traceback.format_exception(type, value, tb))
        lexer     = get_lexer_by_name("pytb", stripall=True)
        formatter = TerminalFormatter()
        if display_type == 'colored':
            if output == 'stderr':
                sys.stderr.write("\n" + highlight(tbtext, lexer, formatter))
            else:
                print("\n" + highlight(tbtext, lexer, formatter))
        elif display_type == 'plain':
            if output == 'stderr':
                sys.stderr.write("\n" + tbtext)
            else:
                print("\n" + tbtext)
        else:
            print("")
            print(f"Error in exception_hook(): Invalid display_type [{display_type}]")

        if pause_at_terminate:
            input("Press ENTER to terminate>> ")

    def redirect_exception(self, hook_func = None, output = 'stdout', display_type = 'colored', pause_at_terminate = True):
        if not use_colored_traceback:
            return

        if hook_func is None:
            hook_func = lambda type, value, tb: \
                        self.exception_hook(type, value, tb, output = output, display_type = display_type, pause_at_terminate = pause_at_terminate)

        sys.excepthook = hook_func

    def print_redict(self, *args, **kwargs):
        return self.redirect(*args, **kwargs)

    def print_warning(self, *args, **kwargs):
        return self.print_redirect(*args, console_out = True, **kwargs)

    def print_redirect(self, *args, **kwargs):
        if kwargs.get("console_out", None) and 'stdout' not in self.redirects:
            del kwargs['console_out']
            self.print_original(*args, **kwargs)
        if kwargs.get("console_out", None) is not None:
            del kwargs['console_out']
        
        for target in self.redirects:
            if type(target) is str:
                if target == 'stdout':
                    self.print_original(*args, **kwargs)

                continue

            fp = target
            if fp is None:
                return

            if kwargs.get('end', None) is None:
                newline = True
            else:
                newline = False
                del kwargs['end']

            for a in args:
                try:
                    fp.write("{}".format(a), **kwargs)
                except:
                    pass

            if newline:
                fp.write("\n")
#                fp.write("\n", **kwargs)

            fp.flush()

    def redict(self, targets = None, mode = 'a', 
            redirect_traceback = True, output_traceback = 'stdout', display_type_traceback = 'colored'):
        return self.redirect(targets = targets, mode = mode,
                            redirect_traceback = redirect_traceback, output_traceback = output_traceback, 
                            display_type_traceback = display_type_traceback)

    def redirect(self, targets = None, mode = 'a', 
            redirect_traceback = True, output_traceback = 'stdout', display_type_traceback = 'colored', pause_at_terminate = True):
        if targets is None:
            for s in self.redirects:
                if s == 'stdout':
                    pass
                elif type(s) is not str:
                    try:
                        s.close()
                    except:
                        pass
            self.redirects = None

        if type(targets) is str:
            targets = [targets]

        builtins.print = self.print_redirect
        for s in targets:
            if s == 'stdout':
                self.redirects.append(s)
            else:
                fp = open(s, mode)
                self.redirects.append(fp)

        if redirect_traceback:
            self.redirect_exception(output = output_traceback, display_type = display_type_traceback, pause_at_terminate = pause_at_terminate)

    def replace_path(self, path = None, template = None, ext_dict = {}):
        if template is None:
            template = os.path.join("{dirname}", "{filebody}-out.txt")
        if path is None:
            path = self.script_fullpath

        if type(template) is list:
            p = []
            for s in template:
                s2 = replace_path(path, s, ext_dict)
                p.append(s2)
            return os.path.join(p[0], *p[1:])
        else:
            return replace_path(path, template, ext_dict)

    def get_inifile_dir(self, use_user_inifile = False,
                        inifile_dirs = ['ini'], create_inidir = False,
                        defval = None):
        if not use_user_inifile:
            return self.script_dir

        for dir in inifile_dirs:
            path = os.path.join(self.script_dir, dir)
            path = os.path.abspath(path)
            if create_inidir or os.path.isdir(path):
                return path

        return defval

    def get_inifile_path(self, inidir, use_user_inifile = False):
        dirname, basename, filebody, ext = split_file_path(self.script_fullpath)
        if not use_user_inifile:
            inifile = os.path.join(inidir, f'{filebody}.ini')
        else:
            inifile = os.path.join(inidir, f'{filebody}_{self.user_nospace}.ini')
        
        inifile = os.path.abspath(inifile)
        
        return inifile

    def add_argument(self, opt, opt_str, var_name = None, desc = '', type = 'str', defval = None, optional = True):
        inf_opt = self.args_opt_inf
        inf_idx = self.args_idx_inf
        if opt is None:
            i = len(inf_idx)
            inf_idx.append( { "opt": i, "opt_str": opt_str, "var_name": var_name, "desc": desc, 
                              "type": type, "defval": defval, "optional": optional } )
        else:
            inf_opt[opt] = { "opt": opt, "opt_str": opt_str, "var_name": var_name, "desc": desc, 
                              "type": type, "defval": defval, "optional": optional }
        return inf_opt, inf_idx

    def read_args(self, vars = None, check_allowed_args = False):
        argv = sys.argv
        args_opt = {}
        args_idx = []
#        print(f"{vars=}")
        if vars is None:
            args_vars = {}
        else:
            args_vars = vars
            
        allowed_keys = self.args_opt_inf.keys()
        i_idx_arg = 0
        for i in range(1, len(argv)):
            arg = del_quote(argv[i])
            m = re.match(r'\s*(\S.*)\s*=\s*(.*)\s*$', arg)
            if m:
                opt = m.groups()[0]
                val = m.groups()[1]
                if check_allowed_args and opt not in allowed_keys:
                    return None, 1, f"Error in tkapplicaiton.read_args(): Invalid argument [{opt}]"
#                    print(f"Error in tkapplicaiton.read_args(): Invalid argument [{opt}]")
#                    exit()

                if opt in allowed_keys:
                    var_type = self.args_opt_inf[opt].get("type", None)
                    if var_type is not None:
                        args_opt[opt] = pconv_by_type(val, var_type, defval = None, strict = True)
                        if args_opt[opt] is None:
                            return None, 2, \
                                f"Error: Argument [{opt}] = [{val}] is not compatible with the required type [{var_type}]"

                    var_name = self.args_opt_inf[opt]["var_name"]
                    if var_name is not None:
#                        print(f"set {var_name=} {args_opt[opt]=}")
                        args_vars[var_name] = args_opt[opt]
                else:
                    args_opt[opt] = val
            else:
                if i_idx_arg < len(self.args_idx_inf):
                    args_idx.append(arg)
                    continue
                else:
                    var_type = self.args_idx_inf[i_idx_arg].get("type", None)
                    if var_type is None:
                        args_idx.append(arg)
                        continue

                val = pconv_by_type(arg, var_type, defval = None, strict = True)
                if val is None:
                    return None, 3, \
                        f"Error: #{i_idx_arg} positional argument = [{arg}] is not compatible with the required type [{var_type}]"

                args_idx.append(val)
                var_name = self.args_idx_inf[i_idx_arg]["var_name"]
                if var_name is not None:
#                    print(f"set {var_name=} {args_opt[opt]=}")
                    args_vars[var_name] = val

                i_idx_arg += 1

        for key in allowed_keys:
            v = self.args_opt_inf[key]
            var = v["var_name"]
            if args_opt.get(key, None) is None:
#                print(f"default {var=} {v['defval']=}")
                args_opt[var] = v["defval"]
                args_vars[var] = v["defval"]

                if not v["optional"]:
                    return None, 4, f"Error: Argument [{key}] is required but not given"
#                print("")
#                print(f"Error: Argument [{key}] is required but not given")
#                exit()

        nposargs = len(args_idx)
        for i in range(len(self.args_idx_inf)):
            v = self.args_idx_inf[i]
            if i >= nposargs:
                var = self.args_idx_inf[i]["var_name"]
                args_vars[var] = v["defval"]
                if not v["optional"]:
                    return None, 5, f"Error: #{i} positional argument is required but not given"
#                print("")
#                print(f"Error: #{i} positional argument is required but not given")
#                exit()
        
        self.args_opt  = args_opt
        self.args_idx  = args_idx
        self.args_vars = args_vars
#        print(f"{args_vars=}")
#        exit()
        
        return args_opt, args_idx, args_vars
    
    def get_arg(self, key, defval = None):
        if type(key) is int:
            try:
                return self.args_idx[key]
            except:
                return defval
        else:
            try:
                return self.args_opt[key]
            except:
                return defval

    def set_usage(self, usage_str = ''):
        self.usage_str   = usage_str

    def usage(self, *args):
        if self.usage_str is not None and self.usage_str != '':
            if self.args_vars:
                s_conv = convert_by_vars(self.usage_str, self.args_vars, prefix = '', bra = r'\{', ket = r'\}')
            else:
                s_conv = self.usage_str

#            print("293")
            for s in s_conv.split('\n'):
                if not re.match(r'\s*[\"\'].*[\"\']', s):
                    s = '"' + s + '"'
                cmd = 'print({})'.format(s.rstrip())
                eval(cmd)
            return

        inf_opt = self.args_opt_inf
        keys_opt = inf_opt.keys()
        inf_idx = self.args_idx_inf
#        print("inf_opt=", inf_opt)
#        print("inf_idx=", inf_idx)
        if len(keys_opt) > 0 or len(inf_idx) > 0:
            usage = f'Usage: python {sys.argv[0]} '
            for key in keys_opt:
                v = inf_opt[key]
                if v['optional']:
                    usage += f"[{v['opt_str']}] "
                else:
                    usage += f"{v['opt_str']} "
            for i in range(len(inf_idx)):
                v = inf_idx[i]
                if v['optional']:
                    usage += f"[{v['opt_str']}] "
                else:
                    usage += f"{v['opt_str']} "

            print(usage)

            print(f"  Options:")
            for key in keys_opt:
                v = inf_opt[key]
                self.print_warning(f"    {v['opt_str']} (def={v['defval']}) (type={v['type']})")
                if v['desc'] != '':
                    print(f"        {v['desc']}")
            for i in range(len(inf_idx)):
                v = inf_idx[i]
                self.print_warning(f"    #{i} {v['opt_str']} (def={v['defval']}) (type={v['type']})")
                if v['desc'] != '':
                    print(f"        {v['desc']}")

            return

    def terminate(self, message = None, usage = None, pause = False):
        if message is not None:
            self.print_warning("")
            self.print_warning(message)

        if usage:
            self.print_warning("")
            usage(self)
        print("")
        
        if pause:
            input("Press ENTER to terminate >> ")
        print("")

        exit()

    def write_resource(app, path, mode):
        with open(f, 'w') as fp:
            for key in app.phrases.keys():
                try:
                    val = app.phrases[key][lang]
                    fp.write(f'{key}:{val}' + '\n')
                except:
                    pass
        return True

    def read_resource(app, path, lang, phrases = None, sep = '='):
        if phrases is None:
            phrases = {}

        fp = open(path, 'r')
        if not fp:
            return phrases

        while 1:
            line = fp.readline()
            if not line:
                break

#            m = re.match(rf'\s*(\S.*)\s*=\s*(.*)\s*\n$', line, flags = re.MULTILINE | re.DOTALL)
            m = re.match(rf'\s*(\S.*)\s*{sep}\s*(.*)\s*\n$', line, flags = re.MULTILINE | re.DOTALL)
            if not m:
                continue
            
            var = m.groups()[0]
            val = m.groups()[1]
            if phrases.get(var, None) is None:
                phrases[var] = {}
            
            phrases[var][lang] = val
#            print(f"{var=} {val=}")

        fp.close()
#        print(f"{phrases=}")
#        exit()
        
        return phrases

    def p(self, str, defval = None, phrases = None):
        if phrases is None:
            phrases = self.phrases

        if str is None or str == '':
            return ''

#        print("")
#        print(f"{str=}")
        s = phrases.get(str, defval)
#        print(f"{s=}")
        if s is None:
            if defval is not None:
                print(f" 336 return defval={defval}")
                return defval
 
#            print(f" return {str=}")
            return str

        s = s.get(self.config.lang, defval)
#        print(f"{s=}")
        if s is None:
            if defval is not None:
#                print(f" 346 return defval={defval}")
                return defval
 
#            print(f" return {str=}")
            return str

#        print(f"last: return {s=}")
        return s

    def print(self, *args, **kwargs):
        print(*args, **kwargs)

    def get_args(self):
        argv = sys.argv
        return argv, len(argv)

    def get_param_dic(self):
        return self.params.__dict__

    def print_attributes(self, print_header = 1):
         keys = self.__dict__.keys()
         if print_header:
             print("attributes in {}".format(self.ClassPath()))
         for key in keys:
            print("  {}: {}".format(key, self.__dict__[key]))

    def print_attributes_warning(self, print_header = 1):
         keys = self.__dict__.keys()
         if print_header:
             self.print_warning("attributes in {}".format(self.ClassPath()))
         for key in keys:
            self.print_warning("  {}: {}".format(key, self.__dict__[key]))

    def print_dict(self, d, header = None):
        print("")
        if header is not None:
            print(header)
        
        if d is None:
            print("  Null dict")
            return

        for key in d.keys():
            print(f"{key}: {d[key]}")

    def print_dict_warning(self, d, header = None):
        print("")
        if header is not None:
            self.print_warning(header)
        
        if d is None:
            self.print_warning("  Null dict")
            return

        for key in d.keys():
            self.print_warning(f"{key}: {d[key]}")

    def load_module(self, module_name, target = '', desc = '', ret_type = 'str', is_print = False):
        try:
            module = import_module(module_name)
        except:
            module = None

        if module is None:
            if is_print:
                print(f"Error in tkapplication.load_module(): Failed to load module [{module_name}]")
            return None
        else:
            setattr(module, 'name', module_name)
            setattr(module, 'target', target)
            setattr(module, 'desc', desc)
            tkmodule = tkModule(name = module_name, module = module)
            self.modules.append({"module": module, "tkModule": tkmodule, "module_name": module_name, "desc": desc})

        if ret_type == 'tkmodule':
            return tkmodule

        if ret_type == 'module':
            return module

        return module_name

    def load_modules(self, plugin_dir = '', fmask = '*.py', target = '', ret_type = 'module', is_print = True):
# Search plug-in files
        files = glob.glob(os.path.join(plugin_dir, fmask))
        module_names = []
        for path in files:
            dirname, basename, filebody, ext = split_file_path(path)
            module_names.append(filebody)

        if plugin_dir != '':
            sys.path.append(plugin_dir)

        modules = []
        module_names2 = []
        for i in range(len(module_names)):
            m = module_names[i]
            module = self.load_module(module_name = m, target = "read_data", desc = m, ret_type = 'module', is_print = False)
            if module:
                modules.append(module)
                module_names2.append(module.name)
                if is_print:
                    print(f"  {m}: loaded")
            else:
                if is_print:
                    print(f"Warning in tkapplication.load_modules(): Invalid module file [{m}.py]")

        return module_names2, modules

    def call(self, module_name, func_name, *args, **kwargs):
        if type(module_name) is not str:
            module_name = module_name.name
#            m = re.search(r"module\s+'(.*?)'", str(module_name))
#            if m:
#                module_name = m.groups()[0]

        ret = None
        for minf in self.modules:
            if minf["module_name"] == module_name:
#                print("module:", minf["module_name"], minf["module"])
                module = minf["module"]
#                print("module names:", module.__dict__.keys())
                
                if func_name not in module.__dict__.keys():
#                    return None
                    pass
                else:
                    ret = module.__dict__[func_name](*args, **kwargs)
                    return ret

#        print(f"call globals.{func_name}")
        if func_name not in self.globals.keys():
            self.terminate(f"Error in tkapplication.call(): Can not find the function [{func_name}] in self.globals\n")
#            print("")
#            print(f"Error in tkapplication.call(): Can not find the function [{func_name}] in self.globals\n")
#            return None
        else:
            ret = self.globals[func_name](*args)

        return ret

    def call_all(self, func_name, *args):
        ncall = 0
        ret = None
        for minf in self.modules:
#            print("module:", minf["module_name"], minf["module"])
            if minf["module"].__dict__.get(func_name, None) is not None:
                ret = minf["module"].__dict__[func_name](*args)
                ncall += 1

        if ncall == 0:
#            print(f"call globals.{func_name}")
            ret = self.globals[func_name](*args)

        return ret

    def get_argv(self):
        return sys.argv, len(sys.argv)

    def get_params(self):
        return self.params

    def get_param_dict(self):
        return self.params.get_param_dict()

    def read_parameters(self, path, merge_params = False, AddSection = 0):
        ini = tkIniFile()
        inf = ini.ReadAll(path, AddSection = AddSection)
        if inf is not None and merge_params:
            self.params += inf

        return inf

    def save_parameters(self, path, params = None, section = 'Parameters', otherparams = None):
        if params is None:
            params = self.get_params()

        ini = tkIniFile(path)
        for i in params.keys():
            val = params.get(key, None)
            if val is not None:
                ini.WriteString(section, key, val)

        if otherparams is not None:
            for key in otherparams.keys():
                val = otherparams.get(key, None)
                if val is not None:
                    ini.WriteString('Preferences', key, val)

    def print_parameters(self, params = None):
        if params is None:
            params = self.get_params()

        for key in params.keys():
            val = params.get(key, None)
            if val is None:
                print("  {:10s}: {:14.8g}".format(key, 'None'))
            else:
                print("  {:10s}: {:14.8g}".format(key, val))


def main():
    print("")
    print("This is library, not runnable")
    print("")


if __name__ == "__main__":
    main()
