import sys
import shlex
import re
from pprint import pprint


from tklib.tkobject import tkObject
from tklib.tkcrystal.tkcifobject import tkCIFObject, tkCIFData
from tklib.tkcrystal.tkcrystal import tkCrystal
from tklib.tkutils import SplitFilePath, split_quoted_args
from tklib.tkfile import tkFile
import tklib.tkre as tkre



"""
Package to manage CIF files
 class tkCIF
"""


#=====================================
# CIF manager class
#=====================================
end_of_cifdata = 0
header    = 1
comment   = 2
singlekey = 3
multikeys = 4
loopkeys  = 5
other     = 999

class tkCIF(tkCIFObject):
    """
    Class to manage CIF files
    """

    def __init__(self, path = None, mode = 'r', **args):
        super(tkCIF, self).__init__(path, mode, **args)
        self.debug = 0
        self.cifdata_list = []
        self.fp   = None
        self.path = None
        self.mode = 'r'
        self.path = self.IfYes(path is not None, path, self.path)
        self.mode = self.IfYes(mode is not None, mode, self.mode)
        self.update(**args)

        if self.path is not None:
            self.Open(self.path, self.mode)
#            print("open [{}]:".format(self.path), self.fp)

    def __del__(self):
        self.Close()
        super(tkCIF, self).__del__()

    def __str__(self):
        return self.ClassPath()


    def splitline(self, lines, i):
        """
        Split lines[i] to a keyword and a value separated by a space
            Input:
                lines: list of given lines
                i: index of lines variable to split
            Return: line, keyword, value
                line: the line associated with keyword and value
                keyword: Non-space keyword
                value: the rest of the line, the header and the tail spaces are striped
        """
        if i >= len(lines):
            return None, '', ''
        match = re.match(r"\s*(\S.*?)\s+(\S.*?)\s*$", lines[i])
        if match:
            return lines[i], match.group(1), match.group(2)
        return lines[i], lines[i].strip(), ''


    def GetNextValuesOne(self, lines, nrequired, i):
        """
        Extract CIF values from one line, lines[i]
            Input:
                lines: list of given lines
                nrequired: number of values required
                i: first index of lines to analyze
            Return: c, status, retline, retval
                c: current index of lines
                status: type of the keyword(s) / status of the last line
                    Stop reading if status == end_of_ci_data or other (error?)
                    See the help for GetNextKey
                retline: line associated with the return values (string)
                retval: values (list)
        """

        l, key, val = self.splitline(lines, i)
# data end
        if l is None:
#            print("dataend")
            return i, end_of_cifdata, '', []

# for single key, single line value
        l0 = l.strip()
        list = tkre.Match(r';\s*(.*)\s*', l0)
        if not list:
            if nrequired == 1:
                self.dprint(1, "GNVOne(single): cA={}  nreq={}  l={}, key={}".format(i, nrequired, l0, key))
                return i+1, singlekey, l0, [l0]
            else:
                self.dprint(1, "GNVOne(mult nreq): cB={}  nreq={}  l={}, key={}".format(i, nrequired, l0, key))
                kws = split_quoted_args(l.strip())
#                try:
#                    kws = shlex.split(l.strip())
#                except:
#                    kws = l.strip()
#                self.dprint(1, "kws=", kws)
                if len(kws) == nrequired:
                    return i+1, multikeys, l0, kws
                return i+1, multikeys, l0, kws

# for single key, multi line value
        if len(list) >= 2:
            s = list[1]
        else:
            s = ''
#        try:
#            s = list[1]
#        except:
#            s = ''
        while 1:
            i += 1
            l, key, val = self.splitline(lines, i)
            self.dprint(1, "GNVOne(mult): cC={}  l={}, key={}".format(i, l0, key))
            l0 = l.strip()
# data end
            if l is None:
                s0 = s.strip()
                if s0 == '':
#                    print("single multi line1")
                    return i, singlekey, s, []
                else:
#                    print("single multi line2")
                    return i, singlekey, s, [s0]
# multi line data end
            list = tkre.Match(';', l0)
            if list:
                s0 = s.strip()
                if s0 == '':
#                    print("single multi line ; 1")
                    return i+1, singlekey, s, []
                else:
#                    print("single multi line ; 2")
                    return i+1, singlekey, s, [s0]
            s += l


    def GetNextValues(self, lines, nrequired, i):
        """
        Extract CIF values from multiple lines starting from lines[i]
            Input:
                lines: list of given lines
                nrequired: number of values required
                i: first index of lines to analyze
            Return: c, status, retline, retval
                c: current index of lines
                status: type of the keyword(s) / status of the last line
                    Stop reading if status == end_of_ci_data or other (error?)
                    See the help for GetNextKey
                retline: Merged lines associated with the return values (string)
                retval: values (list)
        """

        count = 0
        retval = []
        retline = ''
        c = i
        while 1:
            c, type2, l2, values = self.GetNextValuesOne(lines, nrequired, c)
            nv = len(values)
            count += nv
            retval += values
            retline += l2
            self.dprint(1, "GNV: c={}  count={}  type2={}  values={}".format(c, count, type2, values))
            if count >= nrequired:
                return c, multikeys, retline, retval
            if type2 == end_of_cifdata:
                if nv == 0:
                    return c, end_of_cifdata, retline, retval
                else:
                    return c, multikeys, retline, retval


    def GetNextKey(self, lines, i):
        """
        Extract CIF keywords from lines[i]
            Input:
                lines: list of given lines
                i: first index of lines to analyze
            Return: c, status, value
                c: current index of lines
                status: type of the keyword(s) / status of the last line
                    Stop reading if status == end_of_ci_data or other (error?)
                        end_of_cifdata = 0
                        header    = 1
                        comment   = 2
                        singlekey = 3
                        multikeys = 4
                        loopkeys  = 5
                        other     = 999
                keyval: List type of [keyword, value]
        """
# data end
        nlines = len(lines)
        if i >= nlines:
#            print("end_of_cifdata: i >= nlines ({} > {})".format(i, nlines))
            return i, end_of_cifdata, []

        keyinf = []
        c = i
        nlines = len(lines)
        while 1:
            l, key, val = self.splitline(lines, c)
            if l is None:
                break

            self.dprint(1, "GNK: c={}, nlines={}: key=[{}]  val=[{}]".format(c, len(lines), key, val))

            l0 = l.strip()
# null value
            if l0 == '':
                c += 1
                continue
# end of cif
            if re.match(r'#End of', l0):
#                print("end_of_cifdata: #End of")
                return c+1, end_of_cifdata, []
#            if re.match(r'data_', l0):
#                return c, end_of_cifdata, []
# comment
            if re.match('#', l0):
                return c+1, comment, [l0, '']

# first header line
            if re.match('data_', key):
                return c+1, header, ['header', l0]

# for single keyword starting from '_'
            if key == '':
                continue
            elif re.match('_', key):
#                self.dprint(1, "singlekw")
# for single key, multiple line value
                if val == '':
                    c += 1
                    c, type2, l2, values = self.GetNextValues(lines, 1, c)
                    self.dprint(1, "GNK: c3={}  type={}  key={}  val={}".format(c, key, type2, values[0]))
                    return c, singlekey, [key, values[0]]
                else:
                    return c+1, singlekey, [key, val]
# for keyword 'loop_'
            elif key == 'loop_':
                self.dprint(1, "\n")
                self.dprint(1, "======================= loop")
                self.dprint(1, "  ** Read keywords")
# collect keywords starting from '_'
                keywords = []
                c += 1
                while 1:
                    if c >= nlines:
                        break
                    l, key, val = self.splitline(lines, c)
                    if l is None:
                        break
                    self.dprint(1, "GNK: c5={}  key={}  val={}".format(c, key, val))
                    l0 = l.strip()
                    if l0 == '':
                        c += 1
                        continue
#                    if re.match(r"\s*#End of", l) or re.match(r"\s*#=====", l):
                    if re.match(r"\s*#End of", l):
                        c += 1
                        break
#                    if re.match(r"\s*data_", l):
#                        break
                    if re.match(r"\s*#", l):
                        c += 1
                        continue
                    if re.match(r"\s*_", key):
                        keywords.append(key)
                        c += 1
                        continue
                    else:
                        c -= 1
                        break
# collect values untile a new '_' or 'loop_' is found
                c += 1
                nrequired = len(keywords)
                keyvalues = []
                nkeyvalues = 0

                self.dprint(1, "\n")
                self.dprint(1, "  ** Read values (c={}  nlines={}  nrequired={})"
                    .format(c, nlines, nrequired))
                while 1:
                    if c >= nlines:
                        break

                    self.dprint(1, "**check c={} {}".format(c, lines[c]))
                    l0 = lines[c].strip()
#                    print("l0=", l0)
                    if l0 == '':
                        c += 1
                        continue
#                    if re.match(r"\s*#End of", lines[c]) or re.match(r"\s*#=====", lines[c]):
                    if re.match(r"\s*#End of", lines[c]):
                        c += 1
                        break
#                    if re.match(r"\s*data_", lines[c]):
#                        break
                    if re.match(r"\s*#", lines[c]):
                        c += 1
                        continue

                    if re.match("\s*$", lines[c]):
                        c += 1
                        continue

                    if re.match("_", lines[c]) or re.match("loop_", lines[c]):
                        break

                    self.dprint(1, "GNK: for GetNextValues: nreq=", nrequired)
                    c, type2, l, values = self.GetNextValues(lines, nrequired, c)
                    self.dprint(1, "     c4={}  type2={}  key={}  val={}  l={}"
                            .format(c, type2, key, values, l.strip()))

                    if type2 == end_of_cifdata:
                        break
#                    if re.match("_", key) or key == 'loop_':
#                        break

                    nval = len(values)
                    self.dprint(1, "\n")
                    self.dprint(1, "GNK: nrequired={}  nvalues={}  nkeyvalues={}"
                            .format(nrequired, nval, nkeyvalues))
                    self.dprint(1, "  values=", values)        
                    keyval = []
                    for i in range(len(keywords)):
                        self.dprint(1, "GNK: build keyval: [{}]::[{}]"
                            .format(keywords[i], values[i]))
                        keyval.append([keywords[i], values[i]])
                    keyvalues.append(keyval)
                    nkeyvalues += 1

                self.dprint(1, "\n*****************GNK: loop_values break")
                self.dprint(1, "keyvalues:")
                self.dpprint(1, keyvalues)
                self.dprint(1, "")
                return c, loopkeys, keyvalues
            else:
                print(f"Error in tkCIF.GetNextKey(): Invalid key [{l}]")
                return c+1, other, [l, '']
                continue
        return c, end_of_cifdata, []

    def StringsToCIF(self, lines):
        """
        Extract CIF contentis from string list, lines
            Input:
                lines: String list red by ReadNextCIF
            Return:
                cifinf: Dict type: cifinf[keyword] = value
                    Array type of CIF data will be given by the keywords 'keyword[i]',
                    where i is an index starting from 0.
        """
        
        cifinf = {}
        i = 0
        icomment = 0
        while 1:
            self.dprint(1, "S2CIF: i next=", i, "  len(lines)=", len(lines))
            i_next, type, keyinf = self.GetNextKey(lines, i)

            self.dprint(1, "S2CIF: i={} => {}  type=[{}] key=[{}]\n".format(i, i_next, type, keyinf))

            if type == end_of_cifdata or len(keyinf) == 0:
                return cifinf

            self.dprint(1, "keyinf:")
            self.dpprint(1, keyinf)
            self.dprint(1, "")

            if type == end_of_cifdata:
                break
            elif type == comment:
                self.dprint(1, "comment {} [{}]".format(icomment, keyinf[0]))
                cifinf["comment[{}]".format(icomment+1)] = keyinf[0]
                icomment += 1
            elif type == loopkeys:
                for idata in range(len(keyinf)):
                    for j in range(len(keyinf[idata])):
                        cifinf[keyinf[idata][j][0] + "[{}]".format(idata)] = keyinf[idata][j][1]
            else:
                cifinf[keyinf[0]] = keyinf[1]
            i = i_next
        return cifinf


    def ReadNextCIF(self, path = None, print_level = 0):
        """
        Read CIF lines to lines
            path: read CIF from path.
                Input:
                    If path is given, open and close file pointer in this function.
                    If path is not given, read from self.path.
                Return:
                    ret = None: failed to read CIF data
                    lines: String list of CIF data lines
            Side effect:
                If file is already opened, file pointer will not be closed.
                Call self.Close() to close file pointer explicitly.
        """
        if path:
            self.path = path
            if self.Open(self.path) == 0:
                return None
        lines = []
        IsSkip = 1
        ret = None
        iheader = 0
        while 1:
            pos = self.Tell()
            line = self.ReadLine()
#            self.dprint(1, "tkcrystal.tkcif.ReadNextCIF: line: [{}]".format(line.strip()))
            if not line:
                if len(lines) == 0:
                    break
                ret = lines
                break
            if IsSkip and re.match(r'[\r\n]$', line):
                continue
            IsSkip = 0
            if re.match("#End", line):
#                self.dprint(1, "tkcrystal.tkcif.ReadNextCIF: match to end")
                lines.append(line)
                ret = lines
                break
            if re.match("data_", line):
                iheader += 1
                if iheader > 1:
                    self.Seek(pos)
                    ret = lines
                    break
            lines.append(line)

        if path:
            self.Close()
        return ret

    def __ReadCIF1(self, path = None, print_level = 1):
        """
        Helper function to read CIF.
        Only called from ReadCIF.
            Input:
                If find_valid_structure == 0, read the first section only.
                if find_valid_structure == 1, read sections until a valid structure will be found
            Return:
                self.cifdata_list[0] (tkCIFData)
            Side effect:
                self.cifdata_list will be updated
                If file is already opened, file pointer will not be closed.
                Call self.Close() to close file pointer explicitly.
        """

        if path:
            if self.Open(path) == None:
                return []
        elif self.fp:
            self.Rewind()
        else:
            if self.Open(self.path) == None:
                return []
        lines = self.ReadNextCIF()
        if path:
            self.Close()

        dic = self.StringsToCIF(lines)

        cif = tkCIFData()
        cif.path   = self.path
        cif.data   = dic
        cif.lines = lines
        self.cifdata_list = [cif]
        if not cif.CheckCIF():
            exit()

        return self.cifdata_list[0]

    def ReadCIF(self, path = None, find_valid_structure = 1, print_level = 1):
        """
        Read CIF file to tkCIFData list, self.cifdata_list
            Input:
                If find_valid_structure == 0, read the first section only.
                if find_valid_structure == 1, read sections until a valid structure will be found
            Return:
                cifdata (tkCIFData)
        """

        if find_valid_structure:
            cifdatas = self.ReadCIFs(path, print_level = print_level)
            for cifdata in cifdatas:
                lata = cifdata.geterrorf('_cell_length_a')
                if lata is not None and lata > 0:
                    return cifdata
            else:
                return None
        return self.__ReadCIF1(path, print_level = print_level)

    def ReadCIFs(self, path = None, print_level = 1):
        """
        Read a CIF file to tkCIFData list
            Input:
                If path is given, read from path.
                If path is not given, read from self.path.
            Return:
                self.cifdata_list (tkCIFData list)
            Side effect:
                self.cifdata_list will be updated
                If file is already opened, file pointer will not be closed.
                Call self.Close() to close file pointer explicitly.
        """
        if path:
            if self.Open(path) == None:
                return []
        elif self.fp:
            self.Rewind()
        else:
            if self.Open(path) == None:
                return []

        ciflist = []
        while 1:
            lines = self.ReadNextCIF(print_level = print_level)
            if lines == None:
                break
            dic = self.StringsToCIF(lines)
            cif = tkCIFData()
            cif.path   = self.path
            cif.data   = dic
            cif.lines = lines
            ciflist.append(cif)
            self.cifdata_list = ciflist

        if path:
            self.Close()

        return self.cifdata_list


infile = 'test.cif'
#infile = 'test1.cif'
#narg = len(sys.argv)
#if narg >= 2:
#    infile = sys.argv[1]

#debug = 0
#if narg >= 3:
#    debug = int(sys.argv[2])


def main():
    cif = tkCIF(infile)
    cif.debug = debug
    single = 1
    
#    atom = tkAtomType()
#    inf = atom.GetAtomInformation('Au')
#    for key in inf:
#        print("{}: {}".format(key, inf[key]))
#    print("")
#    exit()

    if single:
        cifdata = cif.ReadCIF()
        cifdata.Print()
        cry = cifdata.GetCrystal()
        print("")
        print("==============================================")
        cry.PrintInf()
        print("")
        print("==============================================")

        outfile = 'a.cif'
        cifdata.WriteSimpleCIFFile(outfile)
        
        outfile = 'b.cif'
        cifdata.CreateCIFFileFromCCrystal(cry, outfile)

#        pos0 = [0.3, 0.2, 0.1]
#        isym = 1
#        pos1 = cry.DoSymmetryOperation(pos0, isym)
#        print("pos0 ", pos0, " => sym %d" % (isym), pos1)
    else:
        cifdatas = cif.ReadCIFs()
        for cifdata in cifdatas:
            cifdata.Print()
    cif.Close()


if __name__ == "__main__":
    main()
