import os
import sys
import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt
from matplotlib import pyplot as plt
from matplotlib import gridspec


from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg, analyze_varstr, save_csv
from tklib.tkutils import is_exist, is_file, is_dir
from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile

from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tkmatrix import make_matrix1
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me, Ea_Arrhenius, kayserToHz


"""
Plot ALMODE output
"""


usage_str = '''
f"(i) usage: python {sys.argv[0]} plot fig_type plot_type target kpoint_log tau_dat"
f"           fig_type : 'simple', 'sym', 'fat', 'contour', '3d'"
f"           plot_type: 'linear', 'log'"
f"           target   : 'tau', 'v', 'l', 'kxx', 'kxz', etc"
f"       ex: python {sys.argv[0]} {cparams.mode} {cparams.figtype} {cparams.plottype} {cparams.target} {cparams.kplog_file} {cparams.taudat_file}"
'''[1:-1]



def initialize():
#================================
# Global variables
#================================
    app          = tkApplication(usage_str  = usage_str)
    argv, narg   = app.get_argv()

    cparams             = app.get_params()
    cparams.debug       = 0
    cparams.print_level = 0

    cparams.colors = ['black', 'red', 'blue', 'darkgreen', 'darkorange', 'navy', 
              'slategray', 'hotpink', 'olive', 'chocolate', 'magenta', 
              'green']

    cparams.figsize         = [6, 8]
    cparams.legend_fontsize = 3
    cparams.fontsize        = 16
    cparams.alpha           = 0.5

# mode: 
    cparams.mode    = 'plot'
# figtype: 'simple', 'sym', 'fat', 'contour', '3d'
    cparams.figtype = 'sym'
# plottype: 'linear', 'log'
    cparams.plottype = 'log'
# target: 'tau', 'l', 'v', 'kxx'
    cparams.target  = 'tau'
    cparams.kplog_file  = 'k774.log'
    cparams.taudat_file = 'tau_300K-k774.dat'

    cparams.fatplot_maxsize = 5.0
    cparams.nfmesh   = 301
    cparams.nymesh   = 101
    cparams.wDOS     = 0.02
    cparams.wy_ratio = 0.002
    cparams.ymin_contour = 1.0e-5

    header, ext              = os.path.splitext(app.argv[0])
    filebody                 = os.path.basename(header)
    cparams.parameter_file   = filebody + '.in'
    cparams.parameter_bkfile = filebody + '.in.bak'

    return app, cparams

#=============================
# Treat argments
#=============================
def usage(app):
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def updatevars(app, cparams):
    argv = app.argv
#    if len(argv) <= 1:
#        app.terminate(usage = usage)

    cparams.mode = getarg(1, cparams.mode)
    if cparams.mode == 'plot':
        cparams.figtype     = getarg(2, cparams.figtype)
        cparams.plottype    = getarg(3, cparams.plottype)
        cparams.target      = getarg(4, cparams.target)
        cparams.kplog_file  = getarg(5, cparams.kplog_file)
        cparams.taudat_file = getarg(6, cparams.taudat_file)
    else:
        app.terminate(f"Error in updatevars: Invalid mode [{cparams.mode}]", usage = usage)
        
#    header, ext              = os.path.splitext(app.argv[0])
#    filebody                 = os.path.basename(header)
#    cparams.parameter_file   = filebody + '.in'


def read_kpoints(f):
    fin = tkFile(f, IsPrint = False)
    if not fin.fp:
        app.terminate(f"Error in read_kpoints(): Can not read [{f}]", usage = usage)

    line = fin.skip_to(r"^\s*List of irreducible k points")
#    print("line=", line)

    kp_list = []
    while 1:
        line = fin.readline()
        if not line:
            break
            
        a1 = line.split(':')
        if len(a1) < 2:
            break
        a2 = a1[1].split()
        if len(a2) < 4:
            break

#        print("aa=", *aa)
        kp_list.append({ "ik": pint(a1[0]), "kx": pfloat(a2[0]), "ky": pfloat(a2[1]), "kz": pfloat(a2[2]), "w": pfloat(a2[3]) })

    fin.close()

    return kp_list

def read_taudata(f):
    fin = tkFile(f, IsPrint = False)
    if not fin.fp:
        app.terminate(f"Error in read_taudata(): Can not read [{f}]", usage = usage)

    line = fin.skip_to(r"^#\s+ik,\s+is,\s+Frequency")
#    print("line=", line)

    tau_list = []
    while 1:
        line = fin.readline()
        if not line:
            break
            
        aa = line.split()
        if len(aa) < 16:
            break

#        print("aa=", *aa)
        tau_list.append({ "ik": pint(aa[0]), "is": pint(aa[1]), "f": kayserToHz(pfloat(aa[2])) * 1.0e-12, "tau": pfloat(aa[3]), "v": pfloat(aa[4]), 
                          "l": pfloat(aa[5]), "mult": pint(aa[6]), "kappa": [1.0e6 * pfloat(s) for s in aa[7:]] })

    fin.close()

    return tau_list
    
def plot(app, cparams):
    print("")
    print(f"mode: {cparams.mode}")
    print(f"kpoint log file: {cparams.kplog_file}")
    print(f"tau data file: {cparams.taudat_file}")

    print("")
    print(f"Read k points from {cparams.kplog_file}")
    kp_list = read_kpoints(cparams.kplog_file)
    print("  # of k points: ", len(kp_list))
    print("  K points:")
    for i in range(len(kp_list)):
        l = kp_list[i]
        ik = l["ik"]
        kx = l["kx"]
        ky = l["ky"]
        kz = l["kz"]
        w  = l["w"]
        print(f"    {i:2d}: ik={ik:2d}: k=({kx:12.6f}, {ky:12.6f}, {kz:12.6f}) w={w:12.6f}")

    print("")
    print(f"Read tau data from {cparams.taudat_file}")
    tau_list = read_taudata(cparams.taudat_file)
    """
    print("tau data:")
    for i in range(len(tau_list)):
        l = tau_list[i]
        print("  ik={} is={}: f={} tau={} v={}) MFP={}".format(l["ik"], l["is"], l["f"], l["tau"], l["v"], l["l"]),
              "  kappa=", l["kappa"])
    """

#====================================================
# Plot
#====================================================
    print("")
    print("Plot")
    print(f"Figure type: {cparams.figtype}")
    if (cparams.figtype == 'contour' or cparams.figtype == '3d') and cparams.plottype == 'log':
        print(f"  Plot type [{cparams.plottype}] is not allowed for figure type [{cparams.figtype}]")
        cparams.plottype = 'linear'
        print(f"  Plot type is changed to [{cparams.plottype}]")
        if 'k' not in cparams.target:
            print(f"  Target [{cparams.target}] is not allowed for figure type [{cparams.figtype}]")
            cparams.target = 'kxx'
            print(f"  Target is changed to [{cparams.target}]")
    print(f"Plot type  : {cparams.plottype}")
    print(f"Target: {cparams.target}")
    
    ncolors = len(cparams.colors)

#=========================
# Prepare for target
#=========================
    xx   = []
    yy   = []
    itau = []
    icolor = []
    nk = len(kp_list)
    print("nk=", nk)
    for ik in range(nk):
        xx.append([])
        yy.append([])
        itau.append([])
        icolor.append(ik % ncolors)

    if cparams.target == 'tau':
        key = 'tau'
        ylabel = '$\\tau$ (ps)'
        ylim0  = None
    elif cparams.target == 'v':
        key = 'v'
        ylabel = '$v_g$ (m/s)'
        ylim0  = 1.0e-2
    elif cparams.target == 'l':
        key = 'l'
        ylabel = 'MFP (nm)'
        ylim0  = 1.0e-3
    elif 'k' in cparams.target:
        kindex_dict = {'kxx': 0, 'kxy': 1, 'kxz': 2,
                       'kyx': 3, 'kyy': 4, 'kyz': 5, 
                       'kzx': 6, 'kzy': 7, 'kzz': 8
                      }
        key = 'kappa'
        index = kindex_dict[cparams.target]
        ylabel = '$\kappa_{' + cparams.target.replace('k', '') + '}$ ($\mu$W/m/K)'
        ylim0  = 1.0e-6

    xmax = 0.0
    ymax = 0.0
    for i in range(len(tau_list)):
        l  = tau_list[i]
        ik = l["ik"] - 1
        x  = l["f"]
        if 'k' in cparams.target:
            v = l[key][index]
            if i == 0:
                print(f"** [{index}] is selected for the top figure")
        else:
            v = l[key]
            if i == 0:
                print(f"** {cparams.target} is selected for the top figure")

        if abs(v) > 0.0:
            xx[ik].append(x)
            yy[ik].append(v)
            itau[ik].append(i)
            if xmax < x:
                xmax = x
            if ymax < v:
                ymax = v

#=========================
# Build kappa contour data
#=========================
    fstep = xmax / (cparams.nfmesh - 1)
    if cparams.figtype == 'contour' or cparams.figtype == '3d':
        fstep   = xmax / (cparams.nfmesh - 1)
        ystep   = ymax / (cparams.nymesh - 1)
        wy      = ymax * cparams.wy_ratio
        ifrange = int(4.0 * cparams.wDOS / fstep + 1.0001)
        iyrange = int(4.0 * wy / ystep + 1.0001)
        af      = ifrange / 0.832554611
        ay      = iyrange / 0.832554611

        wtot = 0.0
        for i in range(-ifrange, ifrange + 1):
            for j in range(-iyrange, iyrange + 1):
                wtot += exp( -(i / af)**2 - (j / ay)**2 )

        print("")
        print("Contour parameters")
        print(f"  nfmesh   : {cparams.nfmesh}")
        print(f"  nymesh   : {cparams.nymesh}")
        print(f"  wf       : {cparams.wDOS}")
        print(f"  wy(ratio): {cparams.wDOS}")
        print(f"  wy       : {wy}")
        print(f"  ifrange  : {ifrange}")
        print(f"  iyrange  : {iyrange}")
        print(f"  wtot     : {wtot}")

        f_contour   = [i * fstep for i in range(cparams.nfmesh)]
        y_contour   = [i * ystep for i in range(cparams.nymesh)]
        z_contour   = np.zeros([cparams.nfmesh, cparams.nymesh])
        y_kappa     = np.zeros(cparams.nfmesh)
#        print("dim=", len(f_contour), len(y_contour), len(z_contour), len(z_contour[0]))
        kappamax = 0.0
        for i in range(len(tau_list)):
            l    = tau_list[i]

            if 'k' in cparams.target:
                v = l[key][index]
                kappalabel = '$\kappa_{' + cparams.target.replace('k', '') + '}$'
                if i == 0:
                    print(f"** kappa[{index}] is selected for the contour map")
            else:
                v = l['kappa'][0]
                kappalabel = '$\kappa_{xx}$'
                if i == 0:
                    print(f"** kxx is selected for the contour map")

            f    = l["f"]
            idxf = int(f / fstep)
            idxy = int(v / ystep)
            ik   = l["ik"] - 1
            w    = kp_list[ik]["w"]
#            print("i=", i, f, v, idxf, idxy)

            if kappamax < v:
                kappamax = v

            ifs = max([idxf - ifrange, 0])
            ife = min([cparams.nfmesh-1, idxf + ifrange])
            iys = max([idxy - iyrange, 0])
            iye = min([cparams.nymesh-1, idxy + iyrange])
#            print("i=", i, f, idxf, idxy, ifs, ife, iys, iye, w, v)
            for jf in range(ifs, ife + 1):
                for jy in range(iys, iye + 1):
                    dz = w * v * exp( -((jf - f / fstep) / af)**2 - ((jy - v / ystep) / ay)**2 ) / wtot
                    z_contour[jf][jy] += dz
#                    if dz < 0.0:
#                        print("negative dz: i=", i, f, idxf, idxy, ifs, ife, iys, iye, w, v, wtot)
#                    elif z_contour[jf][jy] < 0.0:
#                        print("negative z: i=", i, f, idxf, idxy, ifs, ife, iys, iye, w, vwtot)

        for jf in range(cparams.nfmesh):
            for jy in range(cparams.nymesh):
                if z_contour[jf][jy] < kappamax * cparams.ymin_contour:
                    z_contour[jf][jy] = 0.0
                else:
                    z_contour[jf][jy] = log(z_contour[jf][jy])
        
#=========================
# Build phonon DOS and kappa spectrum
#=========================
    ifrange = int(4.0 * cparams.wDOS / fstep + 1.0001)
    af = ifrange / 0.832554611
    wtot = 0.0
    for j in range(-ifrange, ifrange + 1):
        wtot += exp( -(j / af)**2 )
    print("DOS parameters")
    print(f"  nfmesh : {cparams.nfmesh}")
    print(f"  wDOS   : {cparams.wDOS}")
    print(f"  ifrange: {ifrange}")
    print(f"  wtot   : {wtot}")
    
    f_DOS   = [i * fstep for i in range(cparams.nfmesh)]
    y_DOS   = np.zeros(cparams.nfmesh)
    y_kappa = np.zeros(cparams.nfmesh)
    kappamax = 0.0
    for i in range(len(tau_list)):
        l    = tau_list[i]
        f    = l["f"]
        idxf = int(f / fstep)
        ik   = l["ik"] - 1
        w    = kp_list[ik]["w"]

        if 'k' in cparams.target:
            v = l[key][index]
            kappalabel = '$\kappa_{' + cparams.target.replace('k', '') + '}$'
            if i == 0:
                print(f"** kappa[{index}] is selected for the middle figure")
        else:
            v = l['kappa'][0]
            kappalabel = '$\kappa_{xx}$'
            if i == 0:
                print(f"** kxx is selected for the middle figure")

        if kappamax < v:
            kappamax = v

        ifs = max([idxf - ifrange, 0])
        ife = min([cparams.nfmesh-1, idxf + ifrange])
#        print("i=", i, f, idxf, ifs, ife, w, v)
        for j in range(ifs, ife + 1):
            coeff = w * exp( -((j - f / fstep) / af)**2 ) / wtot
            y_DOS[j]   += coeff
            y_kappa[j] += v * coeff

#=========================
# Plot
#=========================
    fig = plt.figure(figsize = cparams.figsize)
#    ax1 = fig.add_subplot(1, 1, 1)
    spec = gridspec.GridSpec(ncols = 1, nrows = 3, height_ratios=[4, 1, 1])
    if cparams.figtype == '3d':
        ax1 = fig.add_subplot(spec[0], projection = '3d')
    else:
        ax1 = fig.add_subplot(spec[0])
    ax2 = fig.add_subplot(spec[1], sharex = ax1)
    ax3 = fig.add_subplot(spec[2], sharex = ax1)

    ax1.tick_params(labelbottom = False)
    ax2.tick_params(labelbottom = False)

    if cparams.figtype == 'sym':
        for ik in range(nk):
            l = kp_list[ik]
            kx = l["kx"]
            ky = l["ky"]
            kz = l["kz"]

            color = cparams.colors[icolor[ik]]
            if kx == 0.0 and ky == 0.0 and kz == 0.0:
                marker = 'D'
                markersize = 7.0
            elif kx == ky and ky == kz:
                marker = '^'
                markersize = 3.0
            elif kx == ky or ky == kz or kz == kx:
                marker = 'v'
                markersize = 5.0
            else:
                marker = 'o'
                markersize = 3.0

            line, = ax1.plot(xx[ik], yy[ik], label = f"({kx:6.3f},{ky:6.3f},{kz:6.3f})", linestyle = '', 
                        marker = marker, markersize = markersize, alpha = cparams.alpha,
                        markerfacecolor = color, markeredgecolor = color)
    elif cparams.figtype == 'fat' or cparams.figtype == 'contour':
        for i in range(len(tau_list)):
            tl   = tau_list[i]
            f    = tl["f"]
            if 'k' in cparams.target:
                kv = tl[key][index]
                v  = kv
            else:
                kv = tl['kappa'][0]
                v  = tl[key]
                
            if v == 0.0:
                continue

            marker = 'o'
            markersize = max([cparams.fatplot_maxsize * log(kv) / log(kappamax), 0.2])
#            print("i=", i, f, v, kv, kappamax, markersize)
            color = 'black'
            ax1.plot(f, v, linestyle = '', alpha = cparams.alpha,
                        marker = marker, markersize = markersize, markerfacecolor = color, markeredgecolor = color)

        if cparams.figtype == 'contour':
            ax1.contourf(f_contour, y_contour, z_contour.T, levels = 20, cmap = 'Oranges')
    elif cparams.figtype == '3d':
        X, Y = np.meshgrid(f_contour, y_contour)
        ax1.plot_surface(X, Y, z_contour.T, cmap = 'Oranges')
        ax1.set_xlabel("$f$ (THz)", fontsize = cparams.fontsize)
        ax1.set_zlabel(kappalabel, fontsize = cparams.fontsize)
        ax1.tick_params(labelbottom = True)
    else:
        color = cparams.colors[0]
        marker = 'o'
        markersize = 3.0

        for ik in range(nk):
            line, = ax1.plot(xx[ik], yy[ik], label = f"({kx:6.3f},{ky:6.3f},{kz:6.3f})", linestyle = '', 
                        marker = marker, markersize = markersize, alpha = cparams.alpha, 
                        markerfacecolor = color, markeredgecolor = color)

    ax2.plot(f_DOS, y_kappa)
    ax3.plot(f_DOS, y_DOS)

    if ylim0 is not None:
        ax1.set_ylim([ylim0, ax1.get_ylim()[1]])
    ax1.set_ylabel(ylabel, fontsize = cparams.fontsize)
#    ax1.set_xscale('log')
    if cparams.plottype != 'linear':
        ax1.set_yscale(cparams.plottype)

#    ax1.legend(fontsize = cparams.legend_fontsize, prop={"family":"MS Gothic"}, loc = 'upper left', bbox_to_anchor=(1, 1))
    ax1.tick_params(labelsize = cparams.fontsize)

    ax2.set_ylabel(kappalabel, fontsize = cparams.fontsize)
    ax2.tick_params(labelsize = cparams.fontsize)

    ax3.set_xlabel("$f$ (THz)", fontsize = cparams.fontsize)
    ax3.set_ylabel("DOS", fontsize = cparams.fontsize)
    ax3.tick_params(labelsize = cparams.fontsize)


    annot = ax1.annotate("", xy = (0,0), xytext = (20,20), textcoords = "offset points",
                    bbox = dict(boxstyle = "round", fc = "w"),
                    arrowprops=dict(arrowstyle = "->"))
    annot.set_visible(False)

    def find_nearest_data(x, y):
        minr = 1.0e300
        ihit = None
        for i in range(len(tau_list)):
            l    = tau_list[i]
            f    = l["f"]
            if 'k' in cparams.target:
                v = l[key][index]
            else:
                v = l[key]

            if cparams.plottype == 'log':
                if v > 0.0:
                    r = abs(x - f) / xmax + abs(log(y) - log(v))
                else:
                    r = 1.0e300
            else:
                r = abs(x - f) / xmax + abs(y - v) / ymax
#            print("r=", x, y, f, v, r)
            if minr > r:
                minr = r
                ihit = i
        return ihit
    
    def onclick(event):
        if cparams.figtype == '3d':
            return
        if event.inaxes != ax1:
            return

#        print('%s click: button=%d, x=%d, y=%d, xdata=%f, ydata=%f' %
#              ('double' if event.dblclick else 'single', event.button,
#               event.x, event.y, event.xdata, event.ydata))
        x, y  = event.xdata, event.ydata
        ihit  = find_nearest_data(x, y)
        tl    = tau_list[ihit]
        ik    = tl["ik"]
        _is   = tl["is"]
        f     = tl["f"]
        tau   = tl["tau"]
        v     = tl["v"]
        MFP   = tl["l"]
        kappa = tl["kappa"]
        kl    = kp_list[ik - 1]
        kx    = kl["kx"]
        ky    = kl["ky"]
        kz    = kl["kz"]
        w     = kl["w"]
        c     = w / wtot
        print("")
        print(f"is={_is} ik={ik} q=({kx:6.3f}, {ky:6.3f}, {kz:6.3f}) w={w:6.3f} wtot={wtot:6.3f} coeff=w/wtot={c:10.3g}")
        print(f"    f={f:8.3f} tau={tau:10.3g} v={v:10.3g} MFP={MFP:10.3g}")
        print(f"    kappa=|{kappa[0]:10.3g} {kappa[1]:10.3g} {kappa[2]:10.3g}|  |{c*kappa[0]:10.3g} {c*kappa[1]:10.3g} {c*kappa[2]:10.3g}|")
        print(f"          |{kappa[3]:10.3g} {kappa[4]:10.3g} {kappa[5]:10.3g}|  |{c*kappa[3]:10.3g} {c*kappa[4]:10.3g} {c*kappa[5]:10.3g}|")
        print(f"          |{kappa[6]:10.3g} {kappa[7]:10.3g} {kappa[8]:10.3g}|  |{c*kappa[6]:10.3g} {c*kappa[7]:10.3g} {c*kappa[8]:10.3g}|")

    def hover(event):
        vis = annot.get_visible()
        if event.inaxes == ax2 or event.inaxes == ax3:
            x, y = event.xdata, event.ydata
#            print(f"f={x:8.3f} v={y:10.3g}")
        else:
            cont, ind = ax1.contains(event)
#            cont, ind = line.contains(event)
            x, y = event.xdata, event.ydata
#            print(f"f={x:8.3f} v={y:10.3g}")
#            if cont:
#                update_annot(ind)
#                annot.set_visible(True)
#                fig.canvas.draw_idle()
#            else:
#                if vis:
#                   annot.set_visible(False)
#                   fig.canvas.draw_idle()

# https://sabopy.com/py/matplotlib-69/
    def onpick(event):
#        if event.artist != line:
#            return True

        x, y = event.xdata, event.ydata
        print(f"f={x:8.3f} v={y:10.3g}")

#        N = len(event.ind)
#        if not N:
#            return True
#        thisline = event.artist
#        xdata = thisline.get_xdata()
#        ydata = thisline.get_ydata()
#        ind = event.ind

    fig.canvas.mpl_connect("button_press_event", onclick)
    fig.canvas.mpl_connect("motion_notify_event", hover)
#    fig.canvas.mpl_connect("pick_event", onpick)


    plt.tight_layout()

    use_pause = True
    
    if use_pause:
        plt.pause(0.1)
        print("Press ENTER to exit")
        input()
    else:
        print("")
        print("Close the graph window to terminate")
        plt.show()

    
def main(app, cparams):
    updatevars(app, cparams)

#    print("")
#    print("Read [{}]".format(cparams.parameter_file))
#    app.inf = cparams.read_parameters(cparams.parameter_file)

# check args again to use given parameters
    updatevars(app, cparams)

    print("")
    print( "==========================================================================")
    print( "==========================================================================")
    print(f"     Plot ALMODE output files")
    print( "==========================================================================")
    print( "==========================================================================")
    print("parameter file       : {}".format(cparams.parameter_file))

    print("")
    print("Read [{}]".format(cparams.parameter_file))
    app.inf = cparams.read_parameters(cparams.parameter_file)

# check args again to use given parameters
    updatevars(app, cparams)

    if cparams.mode == 'plot':
        plot(app, cparams)
    else:
        app.terminate(f"Error in main: Invalide mode [{cparams.mode}]", usage = usage)

    print("")
    print("Save parameters to [{}]".format(cparams.parameter_file))
    cparams.save_parameters(cparams.parameter_file, "Preferences")


    app.terminate("", usage = usage)


if __name__ == "__main__":
    app, cparams = initialize()
    main(app, cparams)

