import os
import sys
import shutil
import glob
import csv
import re
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from scipy.interpolate import interp1d
from pprint import pprint
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkvariousdata import tkVariousData
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, pconv, is_list, split_file_path, replace_path, del_quote
from tklib.tkgraphic.tkplotevent import tkPlotEvent


"""
Read SPring-8 HAXPES .txt file
"""

usage_str = '''
"  (i) usage: python {} mode input_file Q2min Q2max Q2step".format(sys.argv[0])
"         ex: python {} {} {} {} {}".format(sys.argv[0], cparams.infile, cparams.xmin, cparams.xmax, cparams.xstep)
'''[1:-1]

plugin_ver = "data:0.2"   # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = '.txt'
extensions  = ['.txt']
input_type  = "SPring-8 HAXPES .txt"
output_type = "MSExcel .xlsx"

figsize = (8, 4)
fontsize = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0
    cparams.findvalidstructure = True

    cparams.infile = f'*{default_ext}'

    cparams.xmin  = '*'
    cparams.xmax  = '*'
    cparams.xstep = '*'


def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate("\nNo argument given", usage = usage)
    
    cparams.infile = getarg(1, cparams.infile)
    cparams.xmin   = getfloatarg(2, cparams.xmin)
    cparams.xmax   = getfloatarg(3, cparams.xmax)
    cparams.xstep  = getfloatarg(4, cparams.xstep)


def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def check_file_type(infile, inf = None, app = None, cparams = None):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != default_ext:
#        print(f"Invalid ext [{ext}] != [{default_ext}]")
        return None

    fp = tkFile(infile, mode = 'r', encoding = None)
    line = fp.readline()
    fp.close()
    
    if line[0:6] == '[Info]':
        return {"file_type": input_type}

    a = line.split()
    if len(a) == 4:
        return {"file_type": "SPring-8 HAXPES simple .txt"}

    return None

def read_simple_text(infile, app = None, cparams = None, is_print = False):
    inf = {}

    fp = tkFile(cparams.infile, mode = 'r', encoding = None)
    line = fp.readline()
    labels = line.split()
    labels = [l.strip() for l in labels]

    inf_list  = [{}, {}]
    data_list = [[], []]
    while 1:
        line = fp.readline()
        if not line:
            break

        a = line.split()
        if len(a) < 4:
            continue

        data_list[0].append(pfloat(a[0]))
        data_list[1].append(pfloat(a[1]))

    dirname, basename, filebody, ext = split_file_path(cparams.infile)
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infile
    inf["sample_name"] = filebody
    inf["yscale"]      = 'linear'
    inf["meta"]        = inf_list
    inf["data_list"]   = [data_list]
    inf["nregion"]     = 1
    inf["nspectrum"]   = 1
    inf["ndata"]       = len(data_list[0])
    inf["labels"]      = [labels]
    for i in range(inf["nregion"]):
        inf_list[i]["sample_name"] = filebody
        inf_list[i]["comment"] = ''

    return inf

def read_data(infile, app = None, cparams = None, is_print = False):
    if is_print:
        print("")
        print(f"Read [{infile}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    ftype = check_file_type(infile, app = app, cparams = cparams)
    if ftype is None:
        return None

    if ftype["file_type"] == "SPring-8 HAXPES simple .txt":
        return read_simple_text(infile, app = None, cparams = cparams, is_print = is_print)

    inf = {}

    fp = tkFile(cparams.infile, mode = 'r', encoding = None)

    fp.skip_to("[Info]")
    while 1:
        pos = fp.tell()
        line = fp.readline()
        if not line:
            break

        if line[0] == '[':
            fp.seek(pos)
            break

        a = line.split('=', 1)
        if len(a) < 2:
            continue

        inf[a[0].strip()] = a[1].strip()
    inf["Number of Regions"] = pint(inf.get("Number of Regions", None))

    inf_list  = []
    data_list = []
    while 1:
        inf1  = {}
        line = fp.skip_to(r"\[Region")
        if not line:
            break

        while 1:
            pos = fp.tell()
            line = fp.readline()
            if not line:
                break

            if line[0] == '[':
                fp.seek(pos)
                break

            a = line.split('=', 1)
            if len(a) < 2:
                continue

            inf1[a[0].strip()] = a[1].strip()

        fp.skip_to(r"\[Info")
        if not line:
            break
        while 1:
            pos = fp.tell()
            line = fp.readline()
            if not line:
                break

            if line[0] == '[':
                fp.seek(pos)
                break

            a = line.split('=', 1)
            if len(a) < 2:
                continue

            inf1[a[0].strip()] = pconv(a[1].strip())

        inf1["sample_name"] = inf1.get('Sample', '')
        inf1["comment"]     = inf1.get('Comments', '')

        x = []
        y = []
        fp.skip_to(r"\[Data")
        if not line:
            break

        c = 0
        while 1:
            pos = fp.tell()
            line = fp.readline()
            if not line:
                if c > 0:
                    inf_list.append(inf1)
                    data_list.append([x, y])
                break

            if line[0] == '[':
                if c > 0:
                    inf_list.append(inf1)
                    data_list.append([x, y])
                fp.seek(pos)
                break

            a = line.split()
            if len(a) < 2:
                continue

            x.append(pfloat(a[0]))
            y.append(pfloat(a[1]))
            c += 1

    fp.close()

    dirname, basename, filebody, ext = split_file_path(cparams.infile)
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infile
    inf["sample_name"] = filebody
    inf["yscale"]      = 'linear'
    inf["meta"]        = inf_list
    inf["data_list"]   = data_list
    inf["nregion"]     = len(data_list)
    inf["labels"]      = []
    for i in range(inf["nregion"]):
        inf["labels"].append(["Energy (eV)", "Intensity"])

    return inf

def save_data(outfiles, inf, app = None, cparams = None, is_print = False):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    if not is_list(labels[0]):
        data_list = [data_list]
        labels    = [labels]

    data_all = []
    labels_all = []
    for d in data_list:
        data_all.extend(d)
        labels_all.extend(['Energy (eV)', 'Intensity'])

    outfile = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels_all, data_all)

    return True

def print_data(inf, app = None, cparams = None):
    labels = inf["labels"]
    print("labels=", labels)
    if not is_list(labels[0]):
        labels = [labels]
    label_str = ''
    for l in labels:
        print("l=", l)
        label_str += "\t".join(l)

    print("")
    print("filename:", inf["filename"])
    print("sample  :", inf["sample_name"])
    print("yscale  :", inf["yscale"])
    print("labels  :", label_str)
    print("nregion :", inf["nregion"])

    meta      = inf["meta"]
    data_list = inf["data_list"]
    nregion   = inf["nregion"]

    print(label_str)
    for k in range(nregion):
        data = data_list[k]
        ndata = len(data[0])
        sample_name = meta[k]['sample_name']
        comments    = meta[k]['comment']
        print(f"Spectrum {k+1}: {sample_name} ({comments})")
        for i in range(ndata):
            print(f"  {data[0][i]:12.4g}\t{data[1][i]:12.4g}")

def plot_data(inf, app = None, cparams = None):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    if not is_list(labels[0]):
        data_list = [data_list]
        labels    = [labels]

    print("")
    print("filename:", inf["filename"])
    print("sample  :", inf["sample_name"])
    print("yscale  :", inf["yscale"])
    print("labels  :", inf["labels"])
    print("nregion :", inf["nregion"])
    
    meta      = inf["meta"]
    labels    = inf["labels"]
    data_list = inf["data_list"]
    nregion = inf["nregion"]

    ncolors = len(colors)
    
    print("")
    print("Plot")

    fig, axes = plt.subplots(1, nregion, figsize = figsize, sharey = True, gridspec_kw = {'wspace': 0})
    if type(axes) is not list and type(axes) is not tuple:
        axes = [axes]

    plt.title(f"{cparams.infile}")
    for i in range(nregion):
        ax = axes[i]
        data = data_list[i]
        x = data[0]
        y = data[1]
        ndata = len(x)

        sample_name = meta[i]['sample_name']
        comments    = meta[i]['comment']
        region_name = meta[i].get('Region Name', f'Region {i+1}')
        print(f"  Spectrum {i+1}: {sample_name}: {region_name} ({comments})")

        ax.tick_params(labelsize = fontsize)
        ax.plot(x, y, label = region_name, color = 'black', linewidth = 0.5)

        ax.invert_xaxis()

    icenter = nregion // 2
    axes[icenter].set_xlabel(labels[0][0], fontsize = fontsize)
    ax.set_ylabel(labels[0][1], fontsize = fontsize)
    ax.legend(fontsize = legend_fontsize)

    plt.tight_layout()
    plt.pause(0.1)
    
    print("")
    input("Press ENTER to terminate>>")

def convert(inf, app = None, cparams = None):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams)
    print_data(inf)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, cparams = cparams)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
