import os
import sys
import shutil
import glob
import csv
import re
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from scipy.interpolate import interp1d
from pprint import pprint
from matplotlib import pyplot as plt

from pymatgen.core.periodic_table import Element
from pymatgen.io.cif import CifParser
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from pymatgen.analysis.diffraction.xrd import XRDCalculator, WAVELENGTHS
from pymatgen.analysis.diffraction.tem import TEMCalculator
from pymatgen.analysis.diffraction.neutron import NDCalculator


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, is_list, split_file_path, replace_path
from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tksci import Gaussian, Lorentzian, GaussLorentz
from tklib.tkcrystal.tkcif import tkCIF
#from tklib.tkcrystal.tkcrystal import tkCrystal
#from tklib.tkcrystal.tkatomtype import tkAtomType
from tklib.tkcrystal.tkxrd import Xray_wavelengths
from tklib.tkgraphic.tkplotevent import tkPlotEvent


"""
Convert CIF file to VASP POSCAR file
"""


plugin_ver = "data:0.2"   # data_list_type = [[x, y]]


# Xray_wavelengths
# URL: https://github.com/materialsproject/pymatgen/blob/v2023.5.10/pymatgen/analysis/diffraction/xrd.py
# 'CuKa', 'CuKa2', 'CuKa1', 'CuKb1', 'MoKa', 'MoKa2', 'MoKa1', 'MoKb1',
# 'CrKa', 'CrKa2', 'CrKa1', 'CrKb1', 'FeKa', 'FeKa2', 'FeKa1', 'FeKb1', 'CoKa', 'CoKa2', 'CoKa1', 'CoKb1',
# 'AgKa', 'AgKa2', 'AgKa1', 'AgKb1'

colors  = ['black', 'red', 'blue', 'darkgreen', 'darkorange', 'hotpink', 'lightgreen', 'cyan', 'yellow', 'magenta', 'chocolate', 
           'navy', 'slategray', 'olive' ]



usage_str = '''
"  (i) usage: python {} mode cif_path wavelength Q2min Q2max Q2step fwhm".format(sys.argv[0])
"         ex: python {} sim {} {} {} {} {} {}".format(sys.argv[0], cparams.infile, cparams.wavelength, cparams.xmin, cparams.xmax, cparams.xstep, cparams.fwhm)
'''[1:-1]



#================================
# global parameters
#================================
input_type  = "CIF"
output_type = "pXRD"

figsize = (8, 4)
fontsize = 12


#def Gaussian(x, x0, whalf):
#    a = whalf / 0.832554611
#    X = (x - x0) / a
#    return exp(-X * X)


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0
    cparams.findvalidstructure = True

    cparams.mode = 'sim'
#    cparams.diffraction_mode = 'xrd'

    cparams.infile = '*.cif'

    cparams.wavelength = "CuKa1*1+CuKa2*0.5"
    cparams.xmin  = 20.0
    cparams.xmax  = 120.0
    cparams.xstep = 0.02
    cparams.fwhm   = 0.2
    cparams.Gfraction = 0.5
    cparams.yscale  = 'linear'


def update_vars(app = None, cparams = None):
    if getarg(2, None) is None:
        app.terminate(usage = usage)
    
    cparams.mode        = getarg(1, cparams.mode)
#    cparams.diffraction_mode = getarg(2, cparams.diffraction_mode)
    cparams.infile      = getarg(2, cparams.infile)
    cparams.wavelength  = getarg(3, cparams.wavelength)
    cparams.xmin       = getfloatarg(4, cparams.xmin)
    cparams.xmax       = getfloatarg(5, cparams.xmax)
    cparams.xstep      = getfloatarg(6, cparams.xstep)
    cparams.fwhm        = getfloatarg(7, cparams.fwhm)
    cparams.Gfraction   = getfloatarg(8, cparams.Gfraction)
    cparams.yscale      = getarg     (9, cparams.yscale)


def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def check_file_type(infile, inf = None, app = None, cparams = None):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != '.cif':
        return None

    return {"file_type": "CIF"}

def read_data(infile, app = None, cparams = None, is_print = False):
    if is_print:
        print("")
        print(f"Read [{infile}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    cif = tkCIF()
    cif.debug = cparams.debug
    cifdata = cif.ReadCIF(infile, find_valid_structure = cparams.findvalidstructure)
    cif.Close()

    if not cifdata:
        if is_print:
            terminate("Error: Could not get cifdat from ciffile [{}]".format(ciffile), usage = usage)
        return None

    cry = cifdata.GetCrystal()
#    cry.PrintInf()

    cif_parser = CifParser(infile)

    dirname, basename, filebody, ext = split_file_path(infile)
    sample_name = cry.SampleName()
    if sample_name == '':
        sample_name = filebody

    inf = {}
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infile
    inf["sample_name"] = sample_name
    inf["yscale"]      = 'linear'
    inf["meta"]        = [] #[inf_list]
    inf["data_list"]   = [] #[data_list]
    inf["nregion"]     = 1 #len(inf["data_list"] )
    inf["labels"]      = []
    for i in range(inf["nregion"]):
        inf["labels"].append([r"2$\theta$ ($\degree$)", "Intensity"])

    inf["cif"] = cif
    inf["cry"] = cry
    inf["cif_parser"] = cif_parser

    return inf

def save_data(outfiles, inf, app = None, cparams = None, print_level = 1):
    x, y = inf["conv_data"]
    outfile = outfiles[0]

    if print_level:
        print("")
        print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, ['2Theta (degree)', 'Intensity'], [x, y])

    return True

def convert(inf, app = None, cparams = None, B = 0.5, fwhm = None, Gfraction = None, print_level = 1):
    if fwhm is None:
        fwhm = cparams.get("fwhm", 0.1)
    if Gfraction is None:
        Gfraction = cparams.get("Gfraction", 0.5)

    cry = inf["cry"]
    parser = inf["cif_parser"]

    labels    = inf["labels"]
    if not is_list(labels[0]):
        labels    = [labels]

    if print_level:
        print(f"X-ray source: {cparams.wavelength}")
    wls = cparams.wavelength.split('+')
    wl_list  = []
    kwl_list = []
    for s in wls:
        a = s.split('*')
        try:
            wl = float(a[0])
        except:
            wl = Xray_wavelengths.get(a[0].strip(), None)

        if wl is None:
            app.terminate(f"Error: Invalid Xray_source: {a[0]}")
        
        if len(a) == 1:
            k = 1.0
        else:
            k = pfloat(a[1])
        wl_list.append(wl)
        kwl_list.append(k)
        print(f"  Source: {a[0]}  Wavelength: {wl} angstrom  Intensity: {k}")

    inf["wl_list"] = wl_list
    inf["kwl_list"] = kwl_list

    nQ2 = int((cparams.xmax - cparams.xmin) / cparams.xstep + 1.000001)
    dQ2 = 12.0 * fwhm
    nconv = int(dQ2 / cparams.xstep + 1.00001)

    if print_level:
#    print(f"Diffraction mode: {cparams.diffraction_mode}")
        print(f"2Theta range: {cparams.xmin:9.5g} - {cparams.xmax:9.5g} at {cparams.xstep:9.5g} degree ({nQ2} points)")
        print(f"2Theta range for a peak: {dQ2:9.5g} degree for {nconv} points")
        print(f"  FWHM of Gaussian functin: {fwhm} degree")

    xmin_diffraction = max([  0.1, cparams.xmin - dQ2])
    xmax_diffraction = min([179.9, cparams.xmax + dQ2])
    x = np.arange(cparams.xmin, cparams.xmax, cparams.xstep)

    mat = parser.get_structures(primitive = False, symmetrized = False)[0]

    wGL = GaussLorentz(0.0, 0.0, fwhm / 2.0, C0 = 1.0, Gfraction = Gfraction, Gwratio = 1.0, A = None)
    xQ2     = np.arange(cparams.xmin, cparams.xmax + cparams.xstep - 1.0e-5, cparams.xstep)
    xrd_cal = np.zeros(nQ2)
    Q2_list   = []
    dhkl_list = []
    hkl_list  = []
    mul_list  = []
    Int_list  = []
    for i in range(len(wl_list)):
        wl = wl_list[i]
        kI = kwl_list[i]

        DW = {}
        atom_types = cry.AtomTypeList()
        for atype in atom_types:
            s = atype.AtomTypeOnly()
            DW[f"Element('{s}')"] = B

#        print("DW=", DW)
        xrd = XRDCalculator(wavelength = wl, debye_waller_factors = DW) #, debye_waller_factors = {Element('Sr'): 0.2, Element('O'): 0.6})
#        xrd = XRDCalculator(wavelength = wl) #, debye_waller_factors = {Element('Sr'): 0.2, Element('O'): 0.6})
        diffractions = xrd.get_pattern(mat, two_theta_range = (xmin_diffraction, xmax_diffraction), scaled = False)

        ndiffractions = len(diffractions.x)
        Q2   = diffractions.x
        dhkl = diffractions.d_hkls
        hkl  = [diffractions.hkls[i][0]['hkl'] for i in range(ndiffractions)]
        mul  = [diffractions.hkls[i][0]['multiplicity'] for i in range(ndiffractions)]
# y includes mul, LP and T
        Int  = [kI * diffractions.y[i] for i in range(ndiffractions)]
        Q2_list.append(Q2)
        dhkl_list.append(dhkl)
        hkl_list.append(hkl)
        mul_list.append(mul)
        Int_list.append(Int)

        if i == 0:
            print("")
            print(f"Diffractions for wl={wl}: {ndiffractions} diffractions")
            print(f"{'h':2} {'k':2} {'l':2}   {'m':2}    {'dhkl':8} {'2Theta':8} {'Intensity'}")

        for id in range(len(Q2)):
            h = hkl[id][0]
            k = hkl[id][1]
            if len(hkl[id]) == 4:
                l = hkl[id][3]
            else:
                l = hkl[id][2]
            print(f"{h:2} {k:2} {l:2}   {mul[id]:2}    {dhkl[id]:8.6f} {Q2[id]:8.6f} {Int[id]:8.4f}")
#        print("  diffractions.hkls[i]=", diffractions.hkls[i])

        print("")
        print(f"Calculate specrum for wl={wl}")
#        xrd_cal = np.zeros(nQ2)
        for idf in range(ndiffractions):
            idx0 = int((Q2[idf] - cparams.xmin) / cparams.xstep)

            for j in range(idx0 - nconv, idx0 + nconv + 1):
                if j < 0 or j >= nQ2:
                    continue

                GL = GaussLorentz(xQ2[j], Q2[idf], fwhm / 2.0, C0 = 1.0, Gfraction = Gfraction, Gwratio = 1.0, A = None)
                xrd_cal[j] += GL / wGL * Int[idf]

    source = []
    Q2     = []
    dhkl   = []
    hkl = []
    mul = []
    Int = []
    for i in range(len(Q2_list)):
        for j in range(len(Q2_list[i])):
            source.append(i)
        Q2.extend(Q2_list[i])
        dhkl.extend(Q2_list[i])
        hkl.extend(hkl_list[i])
        mul.extend(mul_list[i])
        Int.extend(Int_list[i])

    inf["diffractions"] = {"source": source, "Q2": Q2, "dhkl": dhkl, "hkl": hkl, "mul": mul, "intensity": Int}
    inf["conv_data"] = [xQ2, xrd_cal]
    inf["data_list"] = [inf["conv_data"]]

    return inf

def print_data(inf, app = None, cparams = None):
    cry = inf["cry"]
    cry.PrintInf()

def plot_data(inf, app = None, cparams = None, print_level = 1):
    kwl_list = inf["kwl_list"]
    totalI = sum(kwl_list)

    labels    = inf["labels"]
    if not is_list(labels[0]):
        labels    = [labels]

    xQ2, xrd_cal = inf["conv_data"]
    src  = inf["diffractions"]["source"]
    Q2   = inf["diffractions"]["Q2"]
    dhkl = inf["diffractions"]["dhkl"]
    hkl  = inf["diffractions"]["hkl"]
    mul  = inf["diffractions"]["mul"]
    Int  = totalI * np.array(inf["diffractions"]["intensity"])
    ndiffractions = len(Q2)

    fig = plt.figure(figsize = figsize)
    plot_event = tkPlotEvent(plt, distance = 'x')

    ax = fig.add_subplot(111)
    ax.set_title(f"{cparams.infile} - source: {cparams.wavelength}")
    ax.tick_params(labelsize = fontsize)

    line, = ax.plot(xQ2, xrd_cal, label = 'sim', color = 'black', linewidth = 0.5)
#    line, = ax.plot(xQ2, xrd_cal, label = 'sim', picker = 10, color = 'black', linewidth = 0.5)
    data  = ax.plot(Q2, Int, label = 'diffractions', linestyle = '', 
                        marker = 'o', markerfacecolor = 'red', markeredgecolor = 'red', markersize = 1.0)
    for i in range(len(Q2)):
        color = colors[(src[i]+1) % 2]
        ax.plot([Q2[i], Q2[i]], [0.0, Int[i]], linestyle = '-', color = color, linewidth = 0.5)
    ax.axhline(0.0, linestyle = 'dashed', color = 'red', linewidth = 0.5)

#plot_event.add_data({"label": "diffraction", "plot_type": "2D", "axis": ax, "data": [Q2, dhkl, hkl, mul, Int]})
    hkl_str = [f"{hkl[i][0]} {hkl[i][1]} {hkl[i][2]}" for i in range(ndiffractions)]
    plot_event.add_data({"label": "diffraction", "plot_type": "2D", "axis": ax, "data": data,
                                 "x": Q2, 'y': Int, "xlabel": "2Theta", "ylabel": "Intensity",
                                 "xlist": [src, Q2, dhkl,  hkl_str, mul, Int],
                                 "xlabels": ['Xray', 'Q2', 'dhkl', 'hkl', 'multiplicity', 'Intensity']
                                 })

    ax.set_xlim(cparams.xmin, cparams.xmax)

    ax.set_xlabel(labels[0][0], fontsize = fontsize)
    ax.set_ylabel(labels[0][1], fontsize = fontsize)
    if cparams.yscale == 'log':
        ax.set_yscale(cparams.yscale)

    legend = ax.legend()
    legend_labels = [text.get_text() for text in legend.get_texts()]
    legend_objects   = [line, data[0]]
    linewidth_action = [[0.5, 2.0], [None, None]]
    color_action     = [['black', 'black'], ['red', 'green']]
    marker_action    = [[None, None], ['o', 'x']]
    markersize_action      = [[None, None], [1.0, 3.0]]
    markeredgecolor_action = [[None, None], ['red', 'blue']]
    markerfacecolor_action = [[None, None], ['red', 'blue']]

    prev_idx = None
    def on_legend_label_click(event):
        nonlocal prev_idx
        
        if hasattr(event.artist, 'get_text'):
            clicked_label = event.artist.get_text()  # クリックされた凡例ラベルのテキストを取得
        elif hasattr(event.artist, 'get_label'):
            clicked_label = event.artist.get_label()  # クリックされた凡例ラベルのテキストを取得
        else:
            print("no label")
            return

        def flip_status(status):
            if status == 0:
                return 1
            return 0

        def set_graph_attr(line, method, attr):
#            print(method, attr)
            if attr is not None:
                func = getattr(line, method, None)
                if func and callable(func):
                    func(attr)
                else:
                    print("Invalid func")

#        print("clicked_label=", clicked_label)
        for i in range(len(legend_labels)):
            s = legend_labels[i]
            if clicked_label == s:
#                print(f"データ{i}の凡例 {s} がクリックされました")

                line = legend_objects[i]
#                line.view_status = line.__dict__.get('view_status', 1)
                line.view_status = getattr(line, 'view_status', 1)

                set_graph_attr(line, 'set_linewidth',  linewidth_action[i][line.view_status])
                set_graph_attr(line, 'set_color',      color_action[i][line.view_status])
                set_graph_attr(line, 'set_markersize', markersize_action[i][line.view_status])
                set_graph_attr(line, 'set_marker',     marker_action[i][line.view_status])
                set_graph_attr(line, 'set_markerfacecolor', markerfacecolor_action[i][line.view_status])
                set_graph_attr(line, 'set_markeredgecolor', markeredgecolor_action[i][line.view_status])
                fig.canvas.draw()

                prev_idx = i
                line.view_status = flip_status(line.view_status )

# クリックイベントを関連付けます（凡例ラベル用）
#    legend.set_picker(True)
    for line in legend.get_lines():
        line.set_picker(10)
        line.set_pickradius(5)  # クリックを検出する半径を設定
    for text in legend.get_texts():
        text.set_picker(10)
        if hasattr(text, 'set_pickradius'):
            text.set_pickradius(5)  # 新しいmatplotlibのみ対応。クリックを検出する半径を設定

#    fig.canvas.mpl_connect('pick_event', on_legend_label_click)
    plot_event.register_event(fig, event = "pick_event", 
                    callback = on_legend_label_click)

# 機能しない
# 描画更新後にも凡例をクリック可能に保つために再度設定
    def on_draw(event):
#        print("draw")
#        legend.set_picker(True)
        for text in legend.get_texts():
            text.set_picker(True)
        fig.canvas.mpl_connect('pick_event', on_legend_label_click)

#    fig.canvas.mpl_connect('draw_event', on_draw)
    plot_event.register_event(fig, event = "draw_event", 
                    callback = on_draw)

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>")
    

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"  mode: {cparams.mode}")
    print( "Input file:")
    print(f"  CIF file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")
    print(f"X-ray source: {cparams.wavelength}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)


    inf = read_data(cparams.infile, cparams = cparams)
    print_data(inf)
    inf = convert(inf, app = app, cparams = cparams)
    save_data([cparams.outfile], inf, cparams = cparams)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
