import os
import sys
import shutil
import glob
import csv
import re
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from scipy.interpolate import interp1d
from pprint import pprint
from matplotlib import pyplot as plt

from pymatgen.core.periodic_table import Element
from pymatgen.io.cif import CifParser
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from pymatgen.analysis.diffraction.xrd import XRDCalculator, WAVELENGTHS
from pymatgen.analysis.diffraction.tem import TEMCalculator
from pymatgen.analysis.diffraction.neutron import NDCalculator


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, pconv, terminate, split_file_path, replace_path
#from tklib.tkutils import IsDir, IsFile, SplitFilePath
from tklib.tkvariousdata import tkVariousData
from tklib.tkgraphic.tkplotevent import tkPlotEvent


"""
Convert TOPAS file to XRD
"""


plugin_ver = "data:0.1"   # data_list_type = [x, y]


usage_str = '''
"  (i) usage: python {} TOPAS_txt Q2min Q2max Q2step fwhm".format(sys.argv[0])
"         ex: python {} {} {} {} {} {}".format(sys.argv[0], cparams.infile, cparams.xmin, cparams.xmax, cparams.xstep, cparams.fwhm)
'''[1:-1]



#================================
# global parameters
#================================
input_type  = "TOPAS .pro"
output_type = "pXRD"

figsize  = (8, 4)
fontsize = 12
legend_fontsize = 12


def Gaussian(x, x0, whalf):
    a = whalf / 0.832554611
    X = (x - x0) / a
    return exp(-X * X)


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0
    cparams.findvalidstructure = True

    cparams.infile = '*.txt'

    cparams.xmin  = 20.0
    cparams.xmax  = 120.0
    cparams.xstep = 0.02
    cparams.fwhm   = 0.2


def update_vars(app = None, cparams = None):
    if getarg(2, None) is None:
        app.terminate(usage = usage)
    
    cparams.infile      = getarg(1, cparams.infile)
    cparams.xmin        = getfloatarg(2, cparams.xmin)
    cparams.xmax        = getfloatarg(3, cparams.xmax)
    cparams.xstep       = getfloatarg(4, cparams.xstep)
    cparams.fwhm        = getfloatarg(5, cparams.fwhm)


def check_file_type(infile, inf = None, app = None, cparams = None):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != '.pro':
        return None

    fp = tkFile(infile, 'r', encoding = 'shift_jis')
    line0 = fp.readline()
    if line0.strip() != "":
        return None

    r_factors = fp.readline()
    if 'r_' not in r_factors:
        return None

    iters = fp.readline()
    if 'iters' not in iters:
        return None

    fp.close()

    return {"file_type": "TOPAS .pro"}

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, is_print = False):
    if is_print:
        print("")
        print(f"Read [{infile}]")
    
    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    inf = {}

    fp = tkFile(infile, 'r', encoding = 'shift_jis')
    line0 = fp.readline()
    s = fp.readline().split(' ')
    for i in range(0, len(s), 2):
        inf[s[i]] = pfloat(s[i+1])
    
    while 1:
        line = fp.readline()
        if line is None or line == '':
            break

        if line[0] == '{':
            break

        a = line.split()
        na = len(a)
        if na == 0:
            continue
        if na == 1:
           inf[a[0]] = ''
        else:
            inf[a[0]] = pconv(a[1], strict = True, defval = a[1])

#    print("inf=", inf)

    a = fp.readline().split()
    xmin  = pfloat(a[0])
    xstep = pfloat(a[1])
    xmax  = pfloat(a[2])
    ndata  = int((xmax - xmin) / xstep + 1.0e-5)

    labels    = ['2Theta', 'Intensity']
    nspectrum = 1
    yscale    = 'linear'
    path = inf.get('xdd', None)
    if path is None:
        sample_name = 'sample'
    else:
        dirname, basename, filebody, ext = split_file_path(path)
        sample_name = filebody

    print("Q2=", xmin, xstep, xmax, ndata)

    data_list = [[], []]
    for i in range(ndata):
        line = fp.readline()
        if not line:
            break
            
        Q2 = xmin + i * xstep
        data_list[0].append(Q2)
        data_list[1].append(pfloat(line))

    fp.close()

    inf["data_list_type"] = "[x, y]"
    inf["meta"]        = []
    inf["filename"]    = infile
    inf["sample_name"] = sample_name
    inf["labels"]      = labels
    inf["yscale"]      = yscale
    inf["data_list"]   = data_list
    inf["nregion"]     = 1
    inf["nspectrum"]   = nspectrum
    inf["ndata"]       = ndata
    inf["xmin"]        = xmin
    inf["xmax"]        = xmax
    inf["xstep"]       = xstep
    
    return inf

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def print_data(inf, app = None, cparams = None):
    print("")
    print("filename :", inf["filename"])
    print("sample   :", inf["sample_name"])
    print("yscale   :", inf["yscale"])
    print("labels   :", inf["labels"])
    print("nspectrum:", inf["nspectrum"])
    print("ndata    :", inf["ndata"])
    print(f"2Q range : {inf['xmin']} - {inf['xmax']} degrees, [{inf['xstep']}] step")
    
    nspectrum = inf["nspectrum"]
    data_list = inf["data_list"]
    print("\t".join(inf["labels"]))
    for i in range(inf["ndata"]):
        v = [f"{data_list[ispectrum][i]}" for ispectrum in range(nspectrum + 1)]
        print("\t".join(v))

def plot_data(inf, app = None, cparams = None):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    xQ2     = data_list[0]
    xlabel  = labels[0]
    y_list  = data_list[1:]
    ylabels = labels[1:]

    fig = plt.figure(figsize = figsize)

    ax = fig.add_subplot(111)
    ax.set_title(f"{cparams.infile}")
    ax.tick_params(labelsize = fontsize)

    ncolors = len(colors)
    for i in range(len(ylabels)):
        ax.plot(xQ2, y_list[i], label = ylabels[i], color = colors[i % ncolors], linewidth = 0.5)

    Qmin = max([inf["xmin"], cparams.xmin])
    Qmax = min([inf["xmax"], cparams.xmax])
    ax.set_xlim(Qmin, Qmax)

    ax.set_xlabel(r'2$\theta$', fontsize = fontsize)
    ax.set_ylabel('Intensity', fontsize = fontsize)
    ax.legend(fontsize = legend_fontsize)

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>")
    

def save_data(outfiles, inf, app = None, cparams = None, is_print = False):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    outfile   = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels, data_list)

    return True

def convert(inf, app = None, cparams = None):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams)
    print_data(inf)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, cparams = cparams)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
