import os
import sys
import shutil
import glob
import csv
import re
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from scipy.interpolate import interp1d
from pprint import pprint
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkvariousdata import tkVariousData
from tklib.tkexcel_db import tkExcelDB
from tklib.tkgraphic.tkplotevent import tkPlotEvent


"""
Convert XLSX file to XRD
"""


plugin_ver = "data:0.1"   # data_list_type = [x, y]


usage_str = '''
"  (i) usage: python {} mode Bi2OS2.xlsx Q2min Q2max Q2step".format(sys.argv[0])
"         ex: python {} {} {} {} {}".format(sys.argv[0], cparams.infile, cparams.xmin, cparams.xmax, cparams.Q2step)
'''[1:-1]



#================================
# global parameters
#================================
input_type  = "MSExcel .xlsx"
output_type = "pXRD"

figsize = (8, 4)
fontsize = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0
    cparams.findvalidstructure = True

    cparams.infile = '*.xlsx'

    cparams.xmin  = 20.0
    cparams.xmax  = 120.0
    cparams.Q2step = 0.02


def update_vars(app = None, cparams = None):
    if getarg(2, None) is None:
        app.terminate(usage = usage)
    
    cparams.infile = getarg(1, cparams.infile)
    cparams.xmin   = getfloatarg(2, cparams.xmin)
    cparams.xmax   = getfloatarg(3, cparams.xmax)
    cparams.Q2step = getfloatarg(4, cparams.Q2step)


def check_file_type(infile, inf = None, app = None, cparams = None):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != '.xlsx':
        return None

    return {"file_type": "MSExcel .xlsx"}

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, is_print = False):
    if is_print:
        print("")
        print(f"Read [{infile}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    fp = tkExcelDB(cparams.infile, table_name = None, OpenFile = True, CloseFile = True, 
                description = 'MS-Excel for XRD', IsPrint = True)
    max_column = fp.max_column()
    max_row    = fp.max_row()

    labels = []
    for j in range(max_column):
        v = fp.get(irow = 1, icol = j + 1)
        labels.append(v)
    
    data_list = np.zeros([max_column, max_row - 1])
    for j in range(1, max_row):
        for i in range(max_column):
            v = pfloat(fp.get(irow = j + 1, icol = i + 1))
            data_list[i][j - 1] = v

    dirname, basename, filebody, ext = split_file_path(cparams.infile)
    inf = {}
    inf["data_list_type"] = "[x, y]"
    inf["meta"]        = []
    inf["filename"]    = infile
    inf["sample_name"] = basename
    inf["labels"]      = labels
    inf["yscale"]      = 'linear'
    inf["data_list"]   = data_list
    inf["nregion"]     = 1
    inf["nspectrum"]   = max_column - 1
    inf["ndata"]       = max_row - 1
    inf["xmin"]        = data_list[0][0]
    inf["xmax"]        = data_list[0][inf["ndata"] - 1]
    inf["Q2step"]      = data_list[0][1] - data_list[0][0]

    return inf

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def print_data(inf, app = None, cparams = None):
    print("")
    print("filename :", inf["filename"])
    print("sample   :", inf["sample_name"])
    print("yscale   :", inf["yscale"])
    print("labels   :", inf["labels"])
    print("nspectrum:", inf["nspectrum"])
    print("ndata    :", inf["ndata"])
    print(f"2Q range : {inf['xmin']} - {inf['xmax']} degrees, {{inf['Q2step']}} step")
    
    nspectrum = inf["nspectrum"]
    data_list = inf["data_list"]
    print("\t".join(inf["labels"]))
    for i in range(inf["ndata"]):
        v = [f"{data_list[ispectrum][i]}" for ispectrum in range(nspectrum + 1)]
        print("\t".join(v))

def plot_data(inf, app = None, cparams = None):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    xQ2 = data_list[0]
    xlabel = labels[0]
    y_list = data_list[1:]
    ylabels = labels[1:]

    fig = plt.figure(figsize = figsize)

    ax = fig.add_subplot(111)
    ax.set_title(f"{cparams.infile}")
    ax.tick_params(labelsize = fontsize)

    ncolors = len(colors)
    for i in range(len(ylabels)):
        ax.plot(xQ2, y_list[i], label = ylabels[i], color = colors[i % ncolors], linewidth = 0.5)

    Qmin = max([inf["xmin"], cparams.xmin])
    Qmax = min([inf["xmax"], cparams.xmax])
    ax.set_xlim(Qmin, Qmax)

    ax.set_xlabel(r'2$\theta$', fontsize = fontsize)
    ax.set_ylabel('Intensity', fontsize = fontsize)
    ax.legend(fontsize = legend_fontsize)

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>")
    

def save_data(outfiles, inf, app = None, cparams = None, is_print = False):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    outfile   = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels, data_list)

    return True

def convert(inf, app = None, cparams = None):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams)
    print_data(inf)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, cparams = cparams)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
