import os
import sys
import re
import numpy as np
from numpy import sqrt, exp, sin, cos, tan, log, log10, pi
import numpy.linalg as LA 
import csv
from matplotlib import pyplot as plt


sys.path.append("c:/Programs/python/lib")
sys.path.append("d:/Programs/python/lib")

from tkargs import tkArgs
from tkutils import pint, pfloat, getarg, getintarg, getfloatarg, IsFile, IsDir
from tkutils import terminate
from tksci.tkoptimize import tkOptimize, ReduceOptParams, RecoverParams
from tkgraphic.tkplot import tkPlot


"""
Fitting temperature dependence of Hall effect to 2 layers model
"""

#===================================
# physical constants
#===================================
pi   = 3.14159265358979323846
pi2  = 2.0 * pi
h    = 6.6260755e-34    # Js";
hbar = 1.05459e-34      # "Js";
c    = 2.99792458e8     # m/s";
e    = 1.60218e-19      # C";
e0   = 8.854418782e-12; # C<sup>2</sup>N<sup>-1</sup>m<sup>-2</sup>";
kB   = 1.380658e-23     # JK<sup>-1</sup>";
me   = 9.1093897e-31    # kg";
R    = 8.314462618      # J/K/mol
a0   = 5.29177e-11      # m";


#===================================
# parameters
#===================================
mode = 'init'
#mode = 'mullsq'
#mode = 'mu'
#mode = 'ne'
#mode = 'Ea'
#mode = 'all'

infile     = 'ZnO-SCAM.csv'
outcsvfile = None
parameter_path     = 'params.prm'
prevparameter_path = 'params_prev.prm'

# Total layer thickness
thickness  = 50.0       # nm
thickness2 = 10.0       # nm

# weight of Ne for mode == 'all'
w_ne = 10.0

# integration range for Ne, Nh in kB*T
kinteg_range = 6.0

# Mobility #1 parameters
nfunc_mu_llsq = 5
mu_varname  = ['a0', 'a1', 'a2', 'a3', 'a4', 'Eb']
mu_unit     = [  "",   "",   "",   "",   "", "eV"]
mu_pi       = [-1.5, -1.0,  0.0,  1.0,  1.5, None]
mu_val      = [ 0.0,  0.0,  0.0,  0.0,  0.0, 0.00]
mu_optid    = [   1,    1,    1,    1,    1,    1]
#mu_optid    = [   1,    0,    1,    0,    1,    1]
mu_dx       = [10.0, 10.0, 10.0, 10.0, 10.0,  0.1]  # range to make an initial simplex
mu_dx       = np.array(mu_dx)

# Ne #1 parameters
Tfitmin = 150.0
Tfitmax = 350.0
ne_varname = ["Ev", "meffh", "Ec", "meffe", "EA",   "NA", "ED",   "ND"]
ne_unit    = ["eV",    "me", "eV",    "me", "eV", "cm-3", "eV", "cm-3"]
ne_val     = [ 0.0,     1.0,  1.1,     0.3, 0.05,    0.0, 1.08, 3.0e17]
ne_optid   = [   0,       0,    0,       0,    0,      0,    1,      1]
ne_dx      = [ 0.1,     0.1,  0.1,     0.1,  0.1, 1.0e18,  0.5, 5.0e19]
ne_dx = np.array(ne_dx)

# Mobility #2 parameters
mu2_varname  = [   'mu2']
mu2_unit     = ["cm2/Vs"]
mu2_val      = [     1.0]
mu2_optid    = [       1]
mu2_dx       = [     1.0]
mu2_dx = np.array(mu2_dx)

# Ne #2 parameters
ne2_varname  = [ 'Ne2']
ne2_unit     = ["cm-3"]
ne2_val      = [1.0e20]
ne2_optid    = [     1]
ne2_dx       = [1.0e20]
ne2_dx = np.array(ne2_dx)

# Arrhenius plot parameters
norder   = 1    # Order of polynomial to linear LSQ fitting
nlsqdata = 3    # Number of data used to linear LSQ fitting

# Mobility fitting configuration
allow_negative_mu   = 1
penalty_negative_mu = 1.0e4
max_mu              = 1.0e10

# Global variables
T_list     = None
mu_list    = None
ne_list    = None
sigma_list = None

Tmax  = None
Tnorm = []
muinv = []

meffe = None
meffh = None
Nc    = None
Nv    = None
Dc0   = None
Dv0   = None

#===================================
# Optimization parameters
#===================================
# Flag to use linear LSQ results as initial parameters of non-linear LSQ for mode == 'mu'
use_llsq_results = 0

# For EF determination
nmaxiter_bisection = 30
eps_bisection      = 1.0e-6

#algorism: simplex, sd, cg, newton, broyden, dfp, bfgs
algorism = 'simplex'
#algorism = 'sd'
#algorism = 'cg'
#algorism = 'newton'
#algorism = 'broyden'
#algorism = 'dfp'
#algorism = 'bfgs'

# for line search '' or 'none'/'newton', 'one', 'simple', 'exact', 'golden', 'armijo'
lsmode = 'armijo'

# for global optimization
nmaxiter       = 1000
tolx           = 1.0e-5
tolf           = 1.0e-5
iprintinterval = 50
print_level    = 1

# for Newton-Raphson method
dump  = 0.3

# line search parameters
ls_nmaxiter = 100

# for lsmode == 'exact'
ls_h = 1.0e-3  # x step to calculate first and second derivatives

# for lsmode == 'golden'
ls_alphaeps = 1.0e-2
ls_xrange = 0.5
#ls_alpha = 0.03
ls_alpha = 1.0

# graph plot parameters
fplot = 1
ngdata = 51
xgmin = -4.0
xgmax =  4.0
ygmin = -4.0
ygmax =  4.0
tsleep = 0.2


#===================================
# figure configuration
#===================================
fontsize        = 12
legend_fontsize = 6
figsize         = (8, 8)


#==============================================
# other functions
#==============================================
def usage():
    print("")
    print("Usage: python {} init infile".format(sys.argv[0]))
    print("   Create parameter file")
    print("   ex: python {} init {}".format(sys.argv[0], infile))
    print("")
    print("Usage: python {} mullsq parameter_file".format(sys.argv[0]))
    print("   Fitting mobility - T data to 1 layer model by linear LSQ")
    print("   ex: python {} mullsq {}".format(sys.argv[0], parameter_path))
    print("")
    print("Usage: python {} mu parameter_file a0 a1 a2 a3 a4 Eb".format(sys.argv[0]))
    print("   Fitting mobility - T data to 1 layer model")
    print("   ex: python {} mu {} {} {} {} {} {} {}".format(sys.argv[0], parameter_path, *mu_val))
    print("")
    print("Usage: python {} Ea parameter_file norder nlsqdata".format(sys.argv[0]))
    print("   Calculate Ea from Arrhenius plot of Ne")
    print("   ex: python {} Ea {} {} {}".format(sys.argv[0], parameter_path, norder, nlsqdata))
    print("")
    print("Usage: python {} ne parameter_file".format(sys.argv[0]))
    print("   Fitting Ne - T data to 1 layer model")
    print("   ex: python {} ne {}".format(sys.argv[0], parameter_path))
    print("")
    print("Usage: python {} all parameter_file".format(sys.argv[0]))
    print("   Fitting to 2 layers model")
    print("   ex: python {} all {}".format(sys.argv[0], parameter_path))

def updatevars():
    global mode
    global infile, outcsvfile, parameter_path, prevparameter_path
    global mu_val
    global norder, nlsqdata

    argv = sys.argv
#    if len(argv) == 1:
#        terminate()

    mode = getarg( 1, mode)

    if mode == 'init':
        infile = getarg( 2, infile)
        header, ext    = os.path.splitext(infile)
        filebody       = os.path.basename(header)
        parameter_path     = filebody + '.prm'
        prevparameter_path = filebody + '_prev.prm'
    elif mode == 'mullsq':
        parameter_path = getarg( 2, parameter_path)
    elif mode == 'mu':
        parameter_path = getarg( 2, parameter_path)
        for i in range(min([len(mu_val), len(argv)])):
            mu_val[i] = getfloatarg(i+3, mu_val[i])
    elif mode == 'Ea':
        parameter_path = getarg( 2, parameter_path)
        norder   = getarg( 3, norder)
        nlsqdata = getarg( 4, nlsqdata)
    elif mode == 'ne':
        parameter_path = getarg( 2, parameter_path)
    elif mode == 'all':
        parameter_path = getarg( 2, parameter_path)
    else:
        terminate("Error in updatevars(): Invalid mode [{}]".format(mode), usage = usage)

    if mode != 'init':
        header, ext        = os.path.splitext(parameter_path)
        filebody           = os.path.basename(header)
        outcsvfile         = filebody + '-out.csv'
        infile             = filebody + '.csv'
        prevparameter_path = filebody + '_prev.prm'

def read_csv(infile, xmin = None, xmax = None, delimiter = ',', defval = None):
    datalist = []
    try:
        infp = open(infile, "r")
    except:
        print("")
        print("Error in read_csv(): Can not read [{}]".format(infile))
        return None, None

    f = csv.reader(infp, delimiter = delimiter)
    labels = next(f)

    nr = len(labels)
    for ir in range(nr):
        datalist.append([])

    for row in f:
        x = pfloat(row[0])
        do_store = 1
        if xmin is not None and x < xmin:
            do_store = 0
        if xmax is not None and xmax < x:
            do_store = 0
        if do_store:
            for ir in range(nr):
                try:
                    datalist[ir].append(pfloat(row[ir]))
                except:
                    datalist[ir].append(defval)

    return labels, datalist

def pause_plot(use_show = 1):
    if use_show:
        print("")
        print("Close graph window to terminate")
        plt.show()
    else:
        plt.pause(0.1)
        print("")
        print("Press ENTER to exit>>", end = '')
        input()

def mlsq_general(x, y, lsqfunc, nallvars, allvars = None, optids = None):
    if allvars is None or optids is None:
        m = nallvars
    else:
        m = 0
        for i in range(nallvars):
            if optids[i]:
                m += 1

    n = len(x)
    Si  = np.empty([m, 1])
    Sij = np.empty([m, m])

    idx1 = 0
    for l in range(0, nallvars):
        if optids[l] == 0:
            continue

        v = 0.0
        for i in range(0, n):
            v += y[i] * lsqfunc(l, x[i])
        Si[idx1, 0] = v
        idx1 += 1

    idxj = 0
    for j in range(0, nallvars):
        if optids[j] == 0:
            continue

        idxl = idxj
        for l in range(j, nallvars):
            if optids[l] == 0:
                continue

            v = 0.0
            for i in range(0, n):
                v += lsqfunc(j, x[i]) * lsqfunc(l, x[i])

            Sij[idxj, idxl] = Sij[idxl, idxj] = v
            idxl += 1

        idxj += 1

    """
    print("Vector and Matrix:")
    print("Si=")
    print(Si)
    print("Sij=")
    print(Sij)
    print("")
    """
    
    ci = np.linalg.inv(Sij) @ Si
    ci = ci.transpose().tolist()
    ci = ci[0]

    ciall = []
    idx = 0
    for i in range(nallvars):
        if optids[i]:
            print("idx=", idx, ci)
            ciall.append(ci[idx])
            idx += 1
        else:
            ciall.append(allvars[i])
    
    return ciall

def build_optimize(key):
    opt = tkOptimize()

    opt.add_otherparameter('thickness',  thickness)
    opt.add_otherparameter('thickness2', thickness2)
    opt.add_otherparameter('w_ne',       w_ne)
    opt.add_otherparameter('Ea', 0.0)

    if key == 'all' or key == 'mullsq':
        for i in range(nfunc_mu_llsq):
            opt.add_parameter(mu_varname[i], mu_val[i], mu_dx[i], mu_optid[i])
    if key == 'all' or key == 'mu':
        for i in range(len(mu_varname)):
            opt.add_parameter(mu_varname[i], mu_val[i], mu_dx[i], mu_optid[i])
#    if key == 'all' or key == 'Ea':
#        opt.add_parameter('Ea', 0.0, 0.1, 0)
    if key == 'all' or key == 'ne':
        for i in range(len(ne_varname)):
            opt.add_parameter(ne_varname[i], ne_val[i], ne_dx[i], ne_optid[i])
    if key == 'all' or key == 'fit':
        for i in range(len(mu2_varname)):
            opt.add_parameter(mu2_varname[i], mu2_val[i], mu2_dx[i], mu2_optid[i])
        for i in range(len(ne2_varname)):
            opt.add_parameter(ne2_varname[i], ne2_val[i], ne2_dx[i], ne2_optid[i])

    return opt

def backup_parameter(path, backuppath):
    if not os.path.exists(path):
        return 1

    if os.path.exists(backuppath):
        os.remove(backuppath)

    print("")
    print("Backup: Parameter file [{}] is renamed to [{}]".format(path, backuppath))
    os.rename(path, backuppath)
    if os.path.exists(backuppath):
        return 1

    terminate("Error in backup_parameter(): Can not rename [{}] to [{}]".format(path, backuppath))

def save_parameters(opt, path, update_params = 0):
    if opt.save_parameters(path, update_params = update_params):
        print("")
        print("Parameters are stored to [{}]".format(parameter_path))
    else:
        print("")
        print("Error: Can not write to [{}]".format(parameter_path))

def read_parameters(opt, path, update_params = 0, AddSection = 0, repeat_input = 0):
    global infile
    global algorism, nmaxiter
    global lsmode, ls_xrange, ls_h, ls_alpha, ls_dump, ls_nmaxiter, ls_alphaeps
    global tolx, tolf, dump
    global print_level, iprintinterval
    global w_ne

    inf = opt.read_parameters(parameter_path, update_params = update_params, AddSection = AddSection)
    if repeat_input:
        print("read_parameters: parameters in [{}]".format(parameter_path))
        for key, val in inf.items():
            print("  {}: {}".format(key, val))

    infile   = inf['datapath']
    algorism = inf['method']
    lsmode   = inf['lsmode']
    nmaxiter  = pint  (inf['nmaxiter'])
    tolx      = pfloat(inf['tolx'])
    tolf      = pfloat(inf['tolf'])
    dump      = pfloat(inf['dump'])
    ls_xrange = pfloat(inf['ls_xrange'])
    ls_h      = pfloat(inf['ls_h'])
    ls_alpha  = pfloat(inf['ls_alpha'])
    ls_dump   = pfloat(inf['ls_dump'])
    ls_nmaxiter    = pint  (inf['ls_nmaxiter'])
    ls_alphaeps    = pfloat(inf['ls_alphaeps'])
    print_level    = pfloat(inf['print_level'])
    iprintinterval = pint  (inf['iprintinterval'])

    thickness  = pfloat(inf['thickness'])
    thickness2 = pfloat(inf['thickness2'])
    w_ne       = pfloat(inf['w_ne'])
    
    return inf

# mu_n_list: [[mu1, n1], [mu2, n2], ...]
def Hall_multi_carriers(mu_n_list):
    n = len(mu_n_list)

    mu2_product = 1.0
    nmu_sum  = 0.0
    nmu2_sum = 0.0
    for i in range(n):
        mu, n = mu_n_list[i]
        mu2_product *= mu * mu
        nmu_sum     += n * mu
        nmu2_sum    += n * mu * mu

    sigma =  e * nmu_sum
    RH    = 1.0 / e * nmu2_sum / nmu_sum / nmu_sum
    ns    = 1.0 / e / RH
    mus   = sigma / e / ns

    return sigma, RH, ns, mus

def Hall_two_layers(mu1, n1, mu2, n2):
    return Hall_multi_carriers([[mu1, n1], [mu2, n2]])
    
    """
    mu12mu22      = mu1 * mu1 * mu2 * mu2
    n1mu1_p_n2mu2 = n1 * mu1 + n2 * mu2
    nmu2diff = n1  * mu1 * mu1 + n2 * mu2 * mu2

    sigma =  e * n1mu1_p_n2mu2

    a4 = n1mu1_p_n2mu2 * n1mu1_p_n2mu2
    RH = 1.0 / e * nmu2diff / a4
    ns  = 1.0 / e / RH
    mus = sigma / e / ns

    return sigma, RH, ns, mus
    """
    
def Hall_two_carriers(mue, ne, muh, nh):
    sigma, RH, ns, mus = Hall_multi_carriers([[mue, -ne], [muh, nh]])
    return 1.0 / sigma, RH, ns


    """
    muh2mue2 = muh * muh * mue * mue
    nhmuhnemue = nh * muh + ne * mue
#    nhmuenemuh = nh * mue + ne * muh
    nmu2diff = nh  * muh * muh - ne * mue * mue

    rhoxx = 1.0 / e / nhmuhnemue

    a4 = nhmuhnemue * nhmuhnemue
    RH = 1 / e * nmu2diff / a4

    return rhoxx, RH, 1.0 / RH / e
    """

def llsqfunc_mu(ifunc, x):
    return 1.0 / pow(x, mu_pi[ifunc])

def cal_mobility_by_Tnorm(ai, pi, Tnorm, Tmax):
    T = Tnorm * Tmax
    invmu = 0.0
    for i in range(nfunc_mu_llsq):
        invmu += ai[i] / pow(Tnorm, pi[i])

    if invmu <= 0.0:
        return max_mu

    mu = 1.0 / invmu
    try:
        mu = exp(-ai[nfunc_mu_llsq] * e / kB / T) * mu
    except:
        pass
    return mu

Ev    = None
Ec    = None
meffe = None
Dc0   = None
meffh = None
Dv0   = None
EA    = None
NA    = None
ED    = None
ND    = None

# meff in me0, Nc in cm^{-3}
def meff2Nc(meff, T):
    return 2.0 * pow(2.0 * pi * meff * me * kB * T / h / h, 1.5) * 1.0e-6    # cm-3

# meff in me0, Dc0 in cm^{-3}eV^{-1.5}
def meff2Dc0(meff):
    return 2.0 * 2.0 * pi * pow(2.0 * meff * me, 1.5) / h / h / h * 1.0e-6 * pow(e, 1.5)

# Nc in cm^{-3}, meff in me0
def Nc2meff(Nc, T):
#NC = 2 * (2 pi m* kT / h / h)^1.5
    meff = Nc * 1.0e6 / 2.0 / pow(2.0 * pi * kB * T / h / h, 1.5)
    meff = pow(meff, 2.0 / 3.0) / me
    return meff

#T = 300.0
#meff = Nc2meff(ne_val[3], T)
#Dc0  = meff2Dc0(meff)
#print("Nc=", ne_val[3], meff, Dc0)
#exit()

def integrate(func, E0, E1, h):
    n = int((E1 - E0) / h + 1.000001)
    h = (E1 - E0) / (n - 1)
    y = [func(E0 + i * h) for i in range(n)]
    
    S = 0.5 * (y[0] + y[n-1]) + sum(y[1:n-1])
    return h * S

def Dc(E):
    if E <= Ec:
        return 0.0 
    D = Dc0 * sqrt(E - Ec)
    return D

def Dv(E):
    if E >= Ev:
        return 0.0 
    D = Dv0 * sqrt(Ev - E)
    return D

def safe_exp(x):
    if x < -100.0:
        return 0.0
    elif x > 100.0:
        return exp(100.0)
    return exp(x)

def fe(E, EF, T):
    if T == 0.0:
        if E < EF:
            return 1.0
        else:
            return 0.0

    expEkT = safe_exp((E - EF) * e / kB / T)
    return 1.0 / (expEkT + 1.0)

def Dcfe(E, EF, T):
#    print("   Dcfe:", E, EF, T)
    return Dc(E) * fe(E, EF, T)

def Dvfh(E, EF, T):
#    print("   Dvfh:", E, EF, T)
    return Dv(E) * (1.0 - fe(E, EF, T))

nintegkT = 10
def Ne(EF, T):
    global Nc, Dc0

    Et = kB * T / e
    if EF - Ec > kinteg_range * Et:
        Ne = 2.0 / 3.0 * Dc0 * pow(EF - Ec, 1.5)
#        print(" degenerated:", EF - Ec, Ne)
    elif T == 0.0:
        Ne = 0.0
#        print(" T = 0:", EF - Ec, Ne)
    elif Ec - EF > kinteg_range * Et:
        ee = (Ec - EF) / Et
        Ne = Nc * safe_exp(-ee)
#        print(" non-degenerated approx:", EF - Ec, Ne)
    else:
        Estep = Et / nintegkT
        E1 = max([Ec + kinteg_range * Et, EF + kinteg_range * Et])
        Ne = integrate(lambda E: Dcfe(E, EF, T), Ec, E1, Estep)
#        print(" non-degenerated integ:", EF - Ec, Ne)

    return Ne

def Nh(EF, T):
    global Nv, Dv0

    Et = kB * T / e
    if Ev - EF > kinteg_range * Et:
        Nh = 2.0 / 3.0 * Dv0 * pow(Ev - EF, 1.5)
    elif T == 0.0:
        Nh = 0.0
    elif EF - Ev > kinteg_range * Et:
        ee = (EF - Ev) / Et
        Nh = Nv * safe_exp(-ee)
#        print(" h:non-degenerated approx:", Ev - EF, Nh)
    else:
        Estep = Et / nintegkT
        E0 = min([Ev - kinteg_range * Et, EF - kinteg_range * Et])
        Nh = integrate(lambda E: Dvfh(E, EF, T), E0, Ev, Estep)
#        print(" h:non-degenerated integ:", Ev - EF, Nh, "   Erange=", E0, Ev)

#    print("Nh=", Nh, EF, Ev, Nv, Dv0)
    return Nh

"""
T = 300.0
Nc  = meff2Nc(0.3, T)
Dc0 = meff2Dc0(0.3)
Nv  = meff2Nc(0.3, T)
Dv0 = meff2Dc0(0.3)
Ec = 1.1
Ev = 0.0
print("Nc,Dc0=", Nc, Dc0)
print("Nv,Dv0=", Nv, Dv0)

print("")
print("check fe()")
for i in range(-5, 6):
    dE = i * 0.05
    print("E=", dE, "  fe=", fe(dE, 0.0, T))

print("")
print("check Dc,fe()")
for i in range(-5, 6):
    dE = i * 0.05
    de = Dcfe(Ec + dE, Ec, T)
    dh = Dvfh(Ev - dE, Ev, T)
    print("e:dE=", dE, de, dh, Dc(Ec + dE), fe(Ec + dE, Ec, T))
    print("h:dE=", dE, de, dh, Dv(Ev - dE), 1.0 - fe(Ev - dE, Ev, T))

print("")
print("check ne,nh")
for i in range(-5, 6):
    dE = i * 0.05
    ne = Ne(Ec + dE, T)
    nh = Nh(Ev - dE, T)
    print("dE={:8.3g} ne={:12.4g} nh={:12.4g}".format(dE, ne, nh))
exit()
"""

def NDp(EF, T):
    return ND * (1.0 - fe(ED, EF, T))

def NAm(EF, T):
    return NA * fe(EA, EF, T)

def cal_statistics(T):
#初期範囲として、価電子帯上端と伝導帯下端エネルギーを設定する
    EFmin = Ev - 1.0
    EFmax = Ec + 1.0
# まず、EFmin,EFmaxにおけるΔQを計算し、それぞれが正・負あるいは負・生となっていることを確認する    
    dQmin = Ne(EFmin, T) + NAm(EFmin, T) - Nh(EFmin, T) - NDp(EFmin, T)
    dQmax = Ne(EFmax, T) + NAm(EFmax, T) - Nh(EFmax, T) - NDp(EFmax, T)
    if dQmin * dQmax > 0.0:
        print("Error: Initial Emin and Emax should be chosen as dQmin * dQmax < 0")
        return None, None, None, None, None, None        

# 2分法開始
    for i in range(nmaxiter_bisection):
        EFhalf = (EFmin + EFmax) / 2.0
        Neh    = Ne(EFhalf, T)
        NAmh   = NAm(EFhalf, T)
        Nhh    = Nh(EFhalf, T)
        NDph   = NDp(EFhalf, T)
        dQhalf = Neh + NAmh - Nhh - NDph
#        print("  Iter {}: EFhalf = {:12.8f}  dQhalf = {:12.4g}".format(i, EFhalf, dQhalf))
#        print("     Ne={:10.4e}  Nh={:10.4e}  NA-={:10.4e}  ND+={:10.4e}  dQ={:10.4e}".format(Neh, Nhh, NAmh, NDph, dQhalf))
# EFの精度がepsより小さくなったら敬さん終了
        if abs(EFmin - EFhalf) < eps_bisection and abs(EFmax - EFhalf) < eps_bisection:
#            print("  Success: Convergence reached at EF = {}  dQ = {}".format(EFhalf, dQhalf))
            break
        if dQmin * dQhalf < 0.0:
            EFmax = EFhalf
            dQmax = dQhalf
        else:
            EFmin = EFhalf
            dQmin = dQhalf
    else:
        print("  Failed: Convergence did not reach")
        return None

    return EFhalf, Neh, Nhh, NDph, NAmh, dQhalf


def diff1_ne(i, x, optdata = None):
    return 1.0

def calS2_ne(x, optdata = None):
    global Tmax, T_list, mu_list, ne_list, sigma_list, Tnorm, muinv
    global Ev, Ec, Dv0, Dc0, meffe, meffh, EA, NA, ED, ND

    Ev, meffh, Ec, meffe, EA, NA, ED, ND = x
#    print("p0=", Ev, meffh, Ec, meffe, EA, NA, ED, ND)
#    ai = optdata.RecoverParams(x, optdata.x0_all, optdata.optid)
    S2 = 0.0
    for i in range(len(T_list)):
        T = T_list[i]
        if Tfitmin is not None and T < Tfitmin:
            continue
        if Tfitmax is not None and Tfitmax < T:
            continue

        Nc  = meff2Nc(meffe, T)
        Dc0 = meff2Dc0(meffe)
        Nv  = meff2Nc(meffh, T)
        Dv0 = meff2Dc0(meffh)
        EF, ne, nh, ndp, nam, dQ = cal_statistics(T)
        rhoxx, RH, ns = Hall_two_carriers(1.0, ne, 1.0, nh)
        S2 += (log(abs(ns)) - log(ne_list[i]))**2
    
    return S2


# callback function. return code > 0 indicates error and simplex iteration will be terminated
def callback_ne(optdata):
    x0 = optdata.RecoverParams(optdata.x0, optdata.x0_all, optdata.optid)
    if optdata.iter % iprintinterval == 0:
        print("iter {}:".format(optdata.iter), x0)

    return 0

def exec_ne():
    global T_list, mu_list, ne_list, sigma_list
    global Ev, Ec, Nv, Nc, Dv0, Dc0, meffe, meffh, EA, NA, ED, ND

    print("")
    print("Ne parameter fitting by non-linear LSQ")
    print("parameter file:", parameter_path)

    opt = build_optimize('ne')

    print("")
    print("Read parameters from [{}]".format(parameter_path))
    inf = read_parameters(opt, parameter_path, update_params = 1, AddSection = 0, repeat_input = 0)

    print("")
    print("Data file:", infile)
    print("Output CSV file:", outcsvfile)
    print("Parameter file:", parameter_path)
    print("Backup parameter file:", prevparameter_path)

    print("Read data from [{}]".format(infile))
    labels, datalist = read_csv(infile)
    T_list     = datalist[0]
    mu_list    = datalist[1]
    ne_list    = datalist[2]
    sigma_list = datalist[3]
    ndata = len(mu_list)

    print("")
    print("algorism=", algorism)
    print("lsmode=", lsmode)
    print("")

    print("{:8}  {:12} {:12} {:12}".format('T', 'mu', 'ne', 'sigma'))
    for i in range(ndata):
        print("{:8.4g}  {:12.4g} {:12.4g} {:12.4g}".format(T_list[i], mu_list[i], ne_list[i], sigma_list[i]))

    print("")
    print("==============================")
    print(" Non-linear LSQ by {}, {}".format(algorism, lsmode))
    print("==============================")
    opt.set_functions(fitfunc = None, func = calS2_all, diff1func = None, diff2func = None)
#    opt.set_functions(fitfunc = lambda x: 0.0, func = calS2_mu, diff1func = diff1, diff2func = diff2)
#    opt.initialize(callback = callback_mu)
    opt.initialize(callback = callback_mu)
    var = opt.make_optdata(tsleep = tsleep)

    print("")
    print("Initial parameters")
    opt.print_parameters()
    print("fmin=", inf['fmin'])

    Ev, meffh, Ec, meffe, EA, NA, ED, ND = ne_val
    nesim_list = []
    nhsim_list = []
    nssim_list = []
    print("")
    print("{:>8}  {:>10} {:>12} {:>12} {:>12} {:>12} {:>12} {:>12}".format("T", "EF", "Ne(cm-3)", "Nh(cm-3)", "1/e/$R_H$", "ND+(cm-3)", "NA-(cm-3)", "dQ"))
    for iT in range(ndata):
        T = T_list[iT]
        Nc  = meff2Nc(meffe, T)
        Dc0 = meff2Dc0(meffe)
        Nv  = meff2Nc(meffh, T)
        Dv0 = meff2Dc0(meffh)

        EF, ne, nh, ndp, nam, dQ = cal_statistics(T)
        rhoxx,  RH, ns = Hall_two_carriers(1.0, ne, 1.0, nh)
        print("{:8.3g}  {:10.6g} {:12.6g} {:12.6g} {:12.6g} {:12.6g} {:12.6g} {:12.6g}".format(T, EF, ne, nh, ns, ndp, nam, dQ))
        nesim_list.append(ne)
        nhsim_list.append(nh)
        nssim_list.append(abs(ns))

    opt.set_functions(fitfunc = None, func = calS2_ne, diff1func = None, diff2func = None)
    opt.initialize(callback = callback_all)
    var = opt.make_optdata(tsleep = tsleep)

# Optimize
    xmin, fmin, optdata = opt.optimize()
    ai_ne_nlsq = xmin
    Ev, meffh, Ec, meffe, EA, NA, ED, ND = xmin
    print("")
    print("Parameters optimized by non-linear LSQ:")
    opt.print_parameters(x0 = xmin, f = fmin)

    nefit_list = []
    nhfit_list = []
    nsfit_list = []
    Ne2  = ne_list[0]
    Tmin = T_list[0]
    print("")
    print("{:8}  {:10}  {:12} {:12} {:12} {:12}".format('T', 'EF(eV)', 'Ne(obs)', 'Ne(nlsq)', 'Nh(nlsq)', '1/e/RH'))
    for i in range(ndata):
        T  = T_list[i]
        if Tmin > T:
            Ne2 = ne_list[i]
            Tmin = T

        Nc  = meff2Nc(meffe, T)
        Dc0 = meff2Dc0(meffe)
        Nv  = meff2Nc(meffh, T)
        Dv0 = meff2Dc0(meffh)
        EF, ne, nh, ndp, nam, dQ = cal_statistics(T)
        rhoxx, RH, ns = Hall_two_carriers(1.0, ne, 1.0, nh)
        print("{:8.3g}  {:10.6g} {:12.6g} {:12.6g} {:12.6g} {:12.6g}".format(T, EF, ne_list[i], ne, nh, ns))
        nefit_list.append(ne)
        nhfit_list.append(nh)
        nsfit_list.append(abs(ns))

    print("")
    print("Suggested Ne2 (at lowest T = {} K): {} cm-3".format(Tmin, Ne2))

    opt.set_datapath(infile)
    opt.add_parameter('Ne2', Ne2, ne2_dx[0], ne2_optid[0])
    for i in range(len(ne_varname)):
        opt.add_parameter(ne_varname[i], ai_ne_nlsq[i], ne_dx[i], ne_optid[i])
    save_parameters(opt, parameter_path, update_params = 0)


#=============================
# グラフの表示
#=============================
    fig = plt.figure(figsize = figsize)

    ax1 = fig.add_subplot(1, 2, 1)
    ax2 = fig.add_subplot(1, 2, 2)

    markersize = 10.0
    Ne_range = [min(ne_list) * 0.1, max(ne_list) * 10.0]
    ax1.plot(T_list, ne_list, label = '$N_e$ (obs)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ax1.plot(T_list, nesim_list, label = '$N_e$ (ini)', 
            linestyle = '-', color = 'red', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'red', markeredgecolor = 'red', markersize = markersize / 2.0)
    ax1.plot(T_list, nhsim_list, label = '$N_h$ (ini)', 
            linestyle = '-', color = 'blue', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'blue', markeredgecolor = 'blue', markersize = markersize / 2.0)
    ax1.plot(T_list, nssim_list, label = '$N_h$ (ini)', 
            linestyle = '-', color = 'black', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize / 4.0)
    ax1.set_yscale('log')
    ax1.set_ylim(Ne_range)
    ax1.set_xlabel("T (K)",               fontsize = fontsize)
    ax1.set_ylabel("$N_e$ (cm$^2$/(Vs))", fontsize = fontsize)
    ax1.legend(fontsize = legend_fontsize)
    ax1.tick_params(labelsize = fontsize)
    ax1.tick_params(labelsize = fontsize)

    ax2.plot(T_list, ne_list, label = '$N_e$ (obs)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ax2.plot(T_list, nefit_list, label = '$N_e$ (fin)', 
            linestyle = '-', color = 'red', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'red', markeredgecolor = 'red', markersize = markersize / 2.0)
    ax2.plot(T_list, nhfit_list, label = '$N_h$ (fin)', 
            linestyle = '-', color = 'blue', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'blue', markeredgecolor = 'blue', markersize = markersize / 2.0)
    ax2.plot(T_list, nsfit_list, label = '$1/e/R_H$ (fin)', 
            linestyle = '-', color = 'black', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize / 4.0)
    ax2.set_yscale('log')
    ax2.set_ylim(Ne_range)
    ax2.set_xlabel("T (K)",               fontsize = fontsize)
    ax2.set_ylabel("$N_e$ (cm$^2$/(Vs))", fontsize = fontsize)
    ax2.legend(fontsize = legend_fontsize)
    ax2.tick_params(labelsize = fontsize)
    ax2.tick_params(labelsize = fontsize)

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    pause_plot()


def diff1_mu(i, x, optdata = None):
    return 1.0

def calS2_mu(x, optdata = None):
    global Tmax, T_list, mu_list, ne_list, sigma_list, Tnorm, muinv

    ai = optdata.RecoverParams(x, optdata.x0_all, optdata.optid)
    S2 = 0.0
    for i in range(len(ai)):
        if not allow_negative_mu and ai[i] < 0.0:
            S2 += penalty_negative_mu * ai[i] * ai[i]
#            ai[i] = 0.0

    for i in range(len(Tnorm)):
        mu = cal_mobility_by_Tnorm(ai, mu_pi, Tnorm[i], Tmax)
        S2 += (mu - mu_list[i])**2
    
    return S2


# callback function. return code > 0 indicates error and simplex iteration will be terminated
def callback_mu(optdata):
#    print("callback at iter={}".format(optdata.iter))
    x0 = optdata.RecoverParams(optdata.x0, optdata.x0_all, optdata.optid)
    if optdata.iter % iprintinterval == 0:
        print("iter {}:".format(optdata.iter), x0)

    return 0

def calS2_all(x, optdata = None):
    global Tmax, T_list, mu_list, ne_list, sigma_list, Tnorm, muinv
    global ai, Ev, meffh, meffe, Nv, Dv0, Nc, Dc0, EA, NA, ED, ND, mu2, Ne2

    ai = [0.0] * 6
    ai[0], ai[1], ai[2], ai[3], ai[4], ai[5],  \
        Ev, meffh, Ec, meffe, EA, NA, ED, ND, mu2, ne2 = optdata.RecoverParams(x, optdata.x0_all, optdata.optid)

    S2_mu = 0.0
    S2_ne = 0.0
    """
    allow_negative_mu = 1
    for i in range(len(ai)):
        if not allow_negative_mu and ai[i] < 0.0:
            S2_mu += penalty_negative_mu * ai[i] * ai[i]
#            ai[i] = 0.0
    """
    
    for i in range(len(Tnorm)):
        Tn = Tnorm[i]
        T  = T_list[i]
        Nc  = meff2Nc(meffe, T)
        Dc0 = meff2Dc0(meffe)
        Nv  = meff2Nc(meffh, T)
        Dv0 = meff2Dc0(meffh)

        mu = cal_mobility_by_Tnorm(ai, mu_pi, Tn, Tmax)
        EF, ne, nh, ndp, nam, dQ = cal_statistics(T)
        sigma, RH, ns, mus = Hall_multi_carriers([[mu,   -ne * thickness], 
                                                  [mu,    nh * thickness], 
                                                  [mu2, -ne2 * thickness2]])
        ns /= thickness

        S2_mu += (mus - mu_list[i])**2
        S2_ne += (log(abs(ns)) - log(ne_list[i]))**2

#        print("{:8.4g}  {:12.4g} {:12.4g} {:12.4g} {:12.4g} {:12.4g} {:12.4g}".format(T_list[i], mu_list[i], ne_list[i], mu, ne, mus, ns))

    return S2_mu + w_ne * S2_ne

def callback_all(optdata):
#    print("callback at iter={}".format(optdata.iter))
    x0 = optdata.RecoverParams(optdata.x0, optdata.x0_all, optdata.optid)
    if optdata.iter % iprintinterval == 0:
        print("iter {}:".format(optdata.iter), x0)

    return 0

def exec_all():
    global infile
    global Tmax
    global T_list, mu_list, ne_list, sigma_list
    global Tnorm, muinv
    global mu_val
    global Ev, Ec, meffe, meffh, Nc, Nv, Dc0, Dv0, EA, NA, ED, ND

    print("")
    print("Mobility and Ne parameter fitting to 2 layers model by non-linear LSQ")
    print("parameter file:", parameter_path)

    opt = build_optimize('all')

    print("")
    print("Read parameters from [{}]".format(parameter_path))
    inf = read_parameters(opt, parameter_path, update_params = 1, AddSection = 0, repeat_input = 0)

    print("")
    print("Data file:", infile)
    print("Output CSV file:", outcsvfile)
    print("Parameter file:", parameter_path)
    print("Backup parameter file:", prevparameter_path)

    print("")
    print("Read data from [{}]".format(infile))
    labels, datalist = read_csv(infile)
    T_list     = datalist[0]
    mu_list    = datalist[1]
    ne_list    = datalist[2]
    sigma_list = datalist[3]
    ndata = len(mu_list)

    print("")
    print("algorism=", algorism)
    print("lsmode=", lsmode)
    print("")

    print("{:8}  {:12} {:12} {:12}".format('T', 'mu', 'ne', 'sigma'))
    for i in range(ndata):
        print("{:8.4g}  {:12.4g} {:12.4g} {:12.4g}".format(T_list[i], mu_list[i], ne_list[i], sigma_list[i]))

    Tmax = max(T_list)
    for i in range(ndata):
        T = T_list[i]
        Tnorm.append(T / Tmax)
        muinv.append(1.0 / mu_list[i])

# ai[5] = Eb
    ai = [0.0] * 6
    ai[0], ai[1], ai[2], ai[3], ai[4], ai[5],  \
        Ev, meffh, Ec, meffe, EA, NA, ED, ND, mu2, ne2 = opt.x0
    Nc  = meff2Nc(meffe, T)
    Dc0 = meff2Dc0(meffe)
    Nv  = meff2Nc(meffh, T)
    Dv0 = meff2Dc0(meffh)
    print("")
    print("Mobility parameters:")
    print("  ai=", ai[:5])
    print("  Eb=", ai[5])
    print("Carrier density parameters:")
    print("  Ev, Ec=", Ev, Ec)
    print("  mh*, me*=", meffh, meffe)
    print("  EA, NA=", EA, NA)
    print("  ED, ND=", ED, ND)
    print("Interface parameters:")
    print("  mu2, n2=", mu2, ne2)

    print("")
    print("Initial properties:")
    mu_ini_list = []
    ne_ini_list = []
    nh_ini_list = []
    ns_ini_list = []
    mus_ini_list = []
    print("{:>8}  {:>12} {:>12} {:>12} {:>12} {:>12} {:>12} {:>12}"
            .format('T', 'mu(obs)', 'ne(obs)', 'mu(nlsq)', 'ne(cm-3)', 'nh(cm-3)', 'mus(cm2/Vs)', 'ns(cm-3)'))
    for i in range(ndata):
        Tn = Tnorm[i]
        T  = Tn * Tmax

        mu_ini = cal_mobility_by_Tnorm(ai, mu_pi, Tn, Tmax)
        mu_ini_list.append(mu_ini)

        EF, ne, nh, ndp, nam, dQ = cal_statistics(T)
        ne_ini_list.append(ne)
        nh_ini_list.append(nh)
        sigma, RH, ns, mus = Hall_multi_carriers([[mu_ini, -ne * thickness], 
                                                  [mu_ini,  nh * thickness], 
                                                  [mu2,   -ne2 * thickness2]])
        ns /= thickness
        ns_ini_list.append(abs(ns))
        mus_ini_list.append(mus)
        print("{:8.4g}  {:12.4g} {:12.4g} {:12.4g} {:12.4g} {:12.4g} {:12.4g} {:12.4g}"
                .format(T_list[i], mu_list[i], ne_list[i], mu_ini_list[i], ne_ini_list[i], nh_ini_list[i], mus, ns))

    print("")
    print("==============================")
    print(" Non-linear LSQ by {}, {}".format(algorism, lsmode))
    print("==============================")
    opt.set_functions(fitfunc = None, func = calS2_all, diff1func = None, diff2func = None)
#    opt.set_functions(fitfunc = lambda x: 0.0, func = calS2_mu, diff1func = diff1, diff2func = diff2)
    opt.initialize(callback = callback_mu)
    var = opt.make_optdata(tsleep = tsleep)

    print("")
    print("Initial parameters for non-linear LSQ:")
    opt.print_parameters()
    print("fmin=", inf['fmin'])
    
# calculate initial values
    xmin, fmin, optdata = opt.optimize()
    print("")
    print("Parameters optimized by non-linear LSQ:")
    opt.print_parameters(x0 = xmin, f = fmin)
    ai_mu_nlsq = opt.RecoverParams(xmin, optdata.x0_all, optdata.optid)

    mufit_nlsq_list = []
    print("")
    print("{:8}  {:12} {:12}".format('T', 'mu(obs)', 'mu(nlsq)'))
    for i in range(ndata):
        Tn = Tnorm[i]
        T  = Tn * Tmax

        mufit_nlsq = cal_mobility_by_Tnorm(ai_mu_nlsq, mu_pi, Tn, Tmax)
        mufit_nlsq_list.append(mufit_nlsq)
        print("{:8.4g}  {:12.4g} {:12.4g}".format(T_list[i], mu_list[i], mufit_nlsq_list[i]))

    opt.set_datapath(infile)
    opt.x0 = xmin
    save_parameters(opt, parameter_path)

    ai[0], ai[1], ai[2], ai[3], ai[4], ai[5],  \
        Ev, meffh, Ec, meffe, EA, NA, ED, ND, mu2, ne2 = xmin
    Nc  = meff2Nc(meffe, T)
    Dc0 = meff2Dc0(meffe)
    Nv  = meff2Nc(meffh, T)
    Dv0 = meff2Dc0(meffh)
    print("")
    print("Final properties:")
    mu_fin_list = []
    ne_fin_list = []
    nh_fin_list = []
    ns_fin_list = []
    mus_fin_list = []
    print("{:>8}  {:>12} {:>12} {:>12} {:>12} {:>12} {:>12} {:>12}"
            .format('T', 'mu(obs)', 'ne(obs)', 'mu(nlsq)', 'ne(cm-3)', 'nh(cm-3)', 'mus(cm2/Vs)', 'ns(cm-3)'))
    for i in range(ndata):
        Tn = Tnorm[i]
        T  = Tn * Tmax

        mu_fin = cal_mobility_by_Tnorm(ai, mu_pi, Tn, Tmax)
        mu_fin_list.append(mu_fin)

        EF, ne, nh, ndp, nam, dQ = cal_statistics(T)
        ne_fin_list.append(ne)
        nh_fin_list.append(nh)
        sigma, RH, ns, mus = Hall_multi_carriers([[mu_fin, -ne * thickness], 
                                                  [mu_fin,  nh * thickness], 
                                                  [mu2,   -ne2 * thickness2]])
        ns /= thickness
        ns_fin_list.append(abs(ns))
        mus_fin_list.append(mus)
        print("{:8.4g}  {:12.4g} {:12.4g} {:12.4g} {:12.4g} {:12.4g} {:12.4g} {:12.4g}"
                .format(T_list[i], mu_list[i], ne_list[i], mu_fin_list[i], ne_fin_list[i], nh_fin_list[i], mus, ns))

#=============================
# グラフの表示
#=============================
    fig = plt.figure(figsize = figsize)

    ax1 = fig.add_subplot(1, 2, 1)
    ax2 = fig.add_subplot(1, 2, 2)

    markersize = 10.0
    ax1.plot(T_list, mu_list, label = '$\mu$ (obs)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ax1.plot(T_list, mu_ini_list, label = '$\mu$ (ini)', 
            linestyle = '-', color = 'blue', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'blue', markeredgecolor = 'blue', markersize = markersize / 2.0)
    ax1.plot(T_list, mus_ini_list, label = '$\mu$ (s,ini)', 
            linestyle = '-', color = 'red', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'red', markeredgecolor = 'red', markersize = markersize / 3.0)
    ax1.plot(T_list, mus_fin_list, label = '$\mu$ (s,fin)', 
            linestyle = '-', color = 'green', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'green', markeredgecolor = 'green', markersize = markersize / 3.0)
    ax1.set_xlabel("T (K)",               fontsize = fontsize)
    ax1.set_ylabel("$\mu$ (cm$^2$/(Vs))", fontsize = fontsize)
    ax1.legend(fontsize = legend_fontsize)
    ax1.tick_params(labelsize = fontsize)
    ax1.tick_params(labelsize = fontsize)

    ax2.plot(T_list, ne_list, label = '$N_e$ (obs)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ax2.plot(T_list, ne_ini_list, label = '$N_e$ (ini)', 
            linestyle = '-', color = 'blue', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'blue', markeredgecolor = 'blue', markersize = markersize / 2.0)
    ax2.plot(T_list, ns_ini_list, label = '$N_e$ (s,ini)', 
            linestyle = '-', color = 'red', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'red', markeredgecolor = 'red', markersize = markersize / 3.0)
    ax2.plot(T_list, ns_fin_list, label = '$N_e$ (s,fin)', 
            linestyle = '-', color = 'green', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'green', markeredgecolor = 'green', markersize = markersize / 3.0)
    ax2.set_xlabel("T (K)",       fontsize = fontsize)
    ax2.set_ylabel("log($N_e$ / cm$^{-3}$))", fontsize = fontsize)
    ax2.set_yscale('log')
    ax2.legend(fontsize = legend_fontsize)
    ax2.tick_params(labelsize = fontsize)
    ax2.tick_params(labelsize = fontsize)

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    print("")
    print("Close graph window to terminate")
    plt.show()
    
    
    """
    plt.pause(0.1)
    print("")
    print("Press ENTER to exit>>", end = '')
    input()
    """


def exec_mu():
    global infile
    global Tmax
    global T_list, mu_list, ne_list, sigma_list
    global Tnorm, muinv
    global mu_val

    print("")
    print("Mobility parameter fitting by non-linear LSQ")
    print("parameter file:", parameter_path)

    opt = build_optimize('mu')

    print("")
    print("Read parameters from [{}]".format(parameter_path))
    inf = read_parameters(opt, parameter_path, update_params = 1, AddSection = 0, repeat_input = 0)

    print("")
    print("Data file:", infile)
    print("Output CSV file:", outcsvfile)
    print("Parameter file:", parameter_path)
    print("Backup parameter file:", prevparameter_path)

    print("")
    print("Read data from [{}]".format(infile))
    labels, datalist = read_csv(infile)
    T_list     = datalist[0]
    mu_list    = datalist[1]
    ne_list    = datalist[2]
    sigma_list = datalist[3]
    ndata = len(mu_list)

    print("")
    print("algorism=", algorism)
    print("lsmode=", lsmode)
    print("")

    print("{:8}  {:12} {:12} {:12}".format('T', 'mu', 'ne', 'sigma'))
    for i in range(ndata):
        print("{:8.4g}  {:12.4g} {:12.4g} {:12.4g}".format(T_list[i], mu_list[i], ne_list[i], sigma_list[i]))

    Tmax = max(T_list)
    for i in range(ndata):
        T = T_list[i]
        Tnorm.append(T / Tmax)
        muinv.append(1.0 / mu_list[i])

    print("")
    print("==============================")
    print(" Non-linear LSQ by {}, {}".format(algorism, lsmode))
    print("==============================")
    opt.set_functions(fitfunc = None, func = calS2_mu, diff1func = None, diff2func = None)
#    opt.set_functions(fitfunc = lambda x: 0.0, func = calS2_mu, diff1func = diff1, diff2func = diff2)
#    opt.initialize(callback = callback_mu)
    opt.initialize(callback = callback_mu)
    var = opt.make_optdata(tsleep = tsleep)

    print("")
    print("Initial parameters for non-linear LSQ:")
    opt.print_parameters()
    print("fmin=", inf['fmin'])
    
# calculate initial values
    xmin, fmin, optdata = opt.optimize()
    print("")
    print("Parameters optimized by non-linear LSQ:")
    opt.print_parameters(x0 = xmin, f = fmin)
    ai_mu_nlsq = opt.RecoverParams(xmin, optdata.x0_all, optdata.optid)

    mufit_nlsq_list = []
    mu2  = mu_list[0]
    Tmin = T_list[0]
    print("")
    print("{:8}  {:12} {:12}".format('T', 'mu(obs)', 'mu(nlsq)'))
    for i in range(ndata):
        Tn = Tnorm[i]
        T  = Tn * Tmax
        if Tmin > T:
            mu2 = mu_list[i]
            Tmin = T

        mufit_nlsq = cal_mobility_by_Tnorm(ai_mu_nlsq, mu_pi, Tn, Tmax)
        mufit_nlsq_list.append(mufit_nlsq)
        print("{:8.4g}  {:12.4g} {:12.4g}".format(T_list[i], mu_list[i], mufit_nlsq_list[i]))

    print("")
    print("Suggested mu2 (at lowest T = {} K): {} cm2/Vs".format(Tmin, mu2))

    opt.set_datapath(infile)
    opt.add_parameter('mu2', mu2, mu2_dx[0], mu2_optid[0])
    for i in range(len(mu_varname)):
        opt.update_parameter(mu_varname[i], ai_mu_nlsq[i])
    save_parameters(opt, parameter_path)

#=============================
# グラフの表示
#=============================
    fig = plt.figure(figsize = figsize)

    ax1 = fig.add_subplot(1, 2, 1)
    ax2 = fig.add_subplot(1, 2, 2)

    markersize = 10.0
    ax1.plot(T_list, mu_list, label = '$\mu$ (obs)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ax1.plot(T_list, mufit_nlsq_list, label = '$\mu$ (non-linear lsq)', 
            linestyle = '-', color = 'blue', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'blue', markeredgecolor = 'blue', markersize = markersize / 3.0)
    ax1.set_xlabel("T (K)",               fontsize = fontsize)
    ax1.set_ylabel("$\mu$ (cm$^2$/(Vs))", fontsize = fontsize)
    ax1.legend(fontsize = legend_fontsize)
    ax1.tick_params(labelsize = fontsize)
    ax1.tick_params(labelsize = fontsize)

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    print("")
    print("Close graph window to terminate")
    plt.show()
    
    
    """
    plt.pause(0.1)
    print("")
    print("Press ENTER to exit>>", end = '')
    input()
    """

def exec_mullsq():
    global infile
    global Tmax
    global T_list, mu_list, ne_list, sigma_list
    global Tnorm, muinv
    global mu_val

    print("")
    print("Mobility parameter fitting by linear LSQ")
    print("parameter file:", parameter_path)

    opt = build_optimize('mullsq')

    print("")
    print("Read parameters from [{}]".format(parameter_path))
    inf = read_parameters(opt, parameter_path, update_params = 0, AddSection = 0)
#    print("parameters in [{}]".format(parameter_path))
#    for key, val in inf.items():
#        print("  {}: {}".format(key, val))

    print("")
    print("Data file:", infile)
    print("Output CSV file:", outcsvfile)
    print("Parameter file:", parameter_path)
    print("Backup parameter file:", prevparameter_path)

    print("")
    print("Read data from [{}]".format(infile))
    labels, datalist = read_csv(infile)
    T_list     = datalist[0]
    mu_list    = datalist[1]
    ne_list    = datalist[2]
    sigma_list = datalist[3]
    ndata = len(mu_list)

    print("{:8}  {:12} {:12} {:12}".format('T', 'mu', 'ne', 'sigma'))
    for i in range(ndata):
        print("{:8.4g}  {:12.4g} {:12.4g} {:12.4g}".format(T_list[i], mu_list[i], ne_list[i], sigma_list[i]))

    Tmax = max(T_list)
    for i in range(ndata):
        T = T_list[i]
        Tnorm.append(T / Tmax)
        muinv.append(1.0 / mu_list[i])

    print("")
    print("==============================")
    print(" Linear LSQ")
    print("==============================")
    ai_mu_llsq = mlsq_general(Tnorm, muinv, llsqfunc_mu, nfunc_mu_llsq, 
                    allvars = mu_val[:nfunc_mu_llsq], optids = mu_optid[:nfunc_mu_llsq])
    print("")
    print("Parameters fitted by linear LSQ:")
    for ifunc in range(nfunc_mu_llsq):
        print(" x^{:<6.2g}: {:12.6g}".format(mu_pi[ifunc], ai_mu_llsq[ifunc]))

    mufit_llsq_list = []
    S2_llsq = 0.0
    print("")
    print("{:8}  {:12} {:12}".format('T', 'mu(obs)', 'mu(lsq)'))
    for i in range(ndata):
        Tn = Tnorm[i]
        T  = Tn * Tmax

        invmu = 0.0
        for ifunc in range(nfunc_mu_llsq):
            invmu += ai_mu_llsq[ifunc] * llsqfunc_mu(ifunc, Tn)
        mu_cal = 1.0 / invmu
        mufit_llsq  = cal_mobility_by_Tnorm(ai_mu_llsq, mu_pi, Tn, Tmax)
        mufit_llsq_list.append(mufit_llsq)
        print("{:8.4g}  {:12.4g} {:12.4g}".format(T_list[i], mu_list[i], mufit_llsq_list[i]))
        S2_llsq += (mu_list[i] - mufit_llsq)**2
    print("  S2={:12.6g}".format(S2_llsq))

    opt.set_datapath(infile)
    opt.fmin = S2_llsq
    for i in range(nfunc_mu_llsq):
        opt.update_parameter(mu_varname[i], ai_mu_llsq[i])
    save_parameters(opt, parameter_path)


#=============================
# グラフの表示
#=============================
    fig = plt.figure(figsize = figsize)

    ax1 = fig.add_subplot(1, 2, 1)
    ax2 = fig.add_subplot(1, 2, 2)

    markersize = 10.0
    ax1.plot(T_list, mu_list, label = '$\mu$ (obs)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ax1.plot(T_list, mufit_llsq_list, label = '$\mu$ (linear lsq)', 
            linestyle = '-', color = 'red', linewidth = 0.5,
            marker = 'o', markerfacecolor = 'red', markeredgecolor = 'red', markersize = markersize / 2.0)
    ax1.set_xlabel("T (K)",               fontsize = fontsize)
    ax1.set_ylabel("$\mu$ (cm$^2$/(Vs))", fontsize = fontsize)
    ax1.legend(fontsize = legend_fontsize)
    ax1.tick_params(labelsize = fontsize)
    ax1.tick_params(labelsize = fontsize)

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    pause_plot()


def exec_Ea():
    global infile
    
    print("")
    print("Linear LSQ parameters for Arrhenius plot")
    print("parameter file:", parameter_path)
    print("norder  :", norder)
    print("nlsqdata:", nlsqdata)

    opt = build_optimize('all')

    print("")
    print("Read parameters from [{}]".format(parameter_path))
    inf = read_parameters(opt, parameter_path, update_params = 1, AddSection = 0)
#    print("parameters in [{}]".format(parameter_path))
#    for key, val in inf.items():
#        print("  {}: {}".format(key, val))

    print("")
    print("Read data from [{}]".format(infile))
    labels, datalist = read_csv(infile)
    T_list     = datalist[0]
    mu_list    = datalist[1]
    ne_list    = datalist[2]
    sigma_list = datalist[3]
    ndata = len(mu_list)

    invT_list = []
    logne_list = []
    for i in range(ndata):
        invT_list.append(1000.0 / T_list[i])
        logne_list.append(log10(ne_list[i]))

# Calculate Ea in meV
    ndslope_list = []
    ndslope_list.append((logne_list[1] - logne_list[0]) / (invT_list[1] - invT_list[0]))
    for i in range(1, ndata-1):
        ndslope_list.append((logne_list[i+1] - logne_list[i-1]) / (invT_list[i+1] - invT_list[i-1]))
    ndslope_list.append((logne_list[ndata-1] - logne_list[ndata-2]) / (invT_list[ndata-1] - invT_list[ndata-2]))
    ndEa_list = []
    for i in range(ndata):
        ndEa_list.append(-2.302585092994046 * ndslope_list[i] / e * kB * 1000.0 * 1000.0)   # in meV


# Save an example Ea [eV] to parameter file. The value is taken from the 2nd highest T data
    opt.set_datapath(infile)
    Ea = ndEa_list[1] * 0.001   # in eV
    print("")
    print("Ea = {} eV".format(Ea))
    opt.add_otherparameter('Ea', Ea)
#    print("Ea = ", opt.get_parameter('Ea'))
    save_parameters(opt, parameter_path)


#=============================
# グラフの表示
#=============================
    fig = plt.figure(figsize = figsize)

    ax1 = fig.add_subplot(2, 2, 1)
    ax2 = fig.add_subplot(2, 2, 2)
    ax3 = fig.add_subplot(2, 2, 4)

    markersize = 10.0

    ax1.plot(T_list, ne_list, label = '$N_e$ (obs)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ax1.set_xlabel("1000/T (K$^{-1}$)",       fontsize = fontsize)
    ax1.set_ylabel("log($N_e$ / cm$^{-3}$))", fontsize = fontsize)
    ax1.set_yscale('log')
    ax1.legend(fontsize = legend_fontsize)
    ax1.tick_params(labelsize = fontsize)
    ax1.tick_params(labelsize = fontsize)

    ax2.plot(invT_list, logne_list, label = 'log($N_e$) (obs)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
#    ax1.plot(T_list, mufit_llsq_list, label = '$\mu$ (linear lsq)', 
#            linestyle = '-', color = 'red', linewidth = 0.5,
#            marker = 'o', markerfacecolor = 'red', markeredgecolor = 'red', markersize = markersize / 2.0)
#    ax1.plot(T_list, mufit_nlsq_list, label = '$\mu$ (non-linear lsq)', 
#            linestyle = '-', color = 'blue', linewidth = 0.5,
#            marker = 'o', markerfacecolor = 'blue', markeredgecolor = 'blue', markersize = markersize / 3.0)
    ax2.set_xlabel("1000/T (K$^{-1}$)",       fontsize = fontsize)
    ax2.set_ylabel("log($N_e$ / cm$^{-3}$))", fontsize = fontsize)
    ax2.legend(fontsize = legend_fontsize)
    ax2.tick_params(labelsize = fontsize)
    ax2.tick_params(labelsize = fontsize)

    ax3.plot(invT_list, ndEa_list, label = '$E_a$ (numerical diff)', 
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ax3.plot(ax3.get_xlim(), [0.0, 0.0], linestyle = 'dashed', color = 'red', linewidth = 0.5)
    ax3.set_xlabel("1000/T (K$^{-1}$)",       fontsize = fontsize)
    ax3.set_ylabel("$E_a$ (meV)", fontsize = fontsize)
    ax3.legend(fontsize = legend_fontsize)
    ax3.tick_params(labelsize = fontsize)
    ax3.tick_params(labelsize = fontsize)

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    pause_plot()


def exec_init():
    print("")
    print("Create parameter file [{}]".format(parameter_path))
    print("infile:", infile)

    opt = build_optimize('all')
    opt.set_datapath(infile)
    opt.set_method(algorism, lsmode)
    opt.initialize(
            nmaxiter = nmaxiter, tolx = tolx, tolf = tolf,
            callback = callback_mu, print_level = print_level, iprintinterval = iprintinterval,
            dump = dump,
            ls_xrange = ls_xrange, ls_h = ls_h, ls_alpha = ls_alpha, ls_dump = dump, 
            ls_nmaxiter = ls_nmaxiter, ls_alphaeps = ls_alphaeps
            )

    backup_parameter(parameter_path, prevparameter_path)
    save_parameters(opt, parameter_path)

    print("")
    print("Test: Read parameters from [{}]".format(parameter_path))
    inf = read_parameters(opt, parameter_path, update_params = 0, AddSection = 0)
    print("parameters in [{}]".format(parameter_path))
    for key, val in inf.items():
        print("{}: {}".format(key, val))


def main():
    print("")
    print("==================================================")
    print(" Analyze Hall - T data using 2 layers model")
    print("==================================================")

    updatevars()

    print("mode:", mode)
    if mode == 'init':
        exec_init()
    elif mode == 'file':
        exec_file()
    elif mode == 'mullsq':
        exec_mullsq()
    elif mode == 'mu':
        exec_mu()
    elif mode == 'Ea':
        exec_Ea()
    elif mode == 'ne':
        exec_ne()
    elif mode == 'all':
        exec_all()

    terminate('', usage = usage)


if __name__ == "__main__":
    main()
