import sys
import os
import re
import numpy as np
from numpy import sqrt, exp, sin, cos, tan, pi
import numpy.linalg as LA 
from matplotlib import pyplot as plt


from tklib.tkvariousdata import tkVariousData
from tklib.tkapplication import tkApplication



"""
Calculate weighted mobility and carrier density from conductivity and Seebeck coefficient
"""

#===================================
# physical constants
#===================================
pi   = 3.14159265358979323846
pi2  = 2.0 * pi
h    = 6.6260755e-34    # Js";
hbar = 1.05459e-34      # "Js";
c    = 2.99792458e8     # m/s";
e    = 1.60218e-19      # C";
e0   = 8.854418782e-12; # C<sup>2</sup>N<sup>-1</sup>m<sup>-2</sup>";
kB   = 1.380658e-23     # JK<sup>-1</sup>";
me   = 9.1093897e-31    # kg";
R    = 8.314462618      # J/K/mol
a0   = 5.29177e-11      # m";


#===================================
# parameters
#===================================
app = None

# mode: 'single', 'X', 'T'
mode = 'single'

# Temperature (K)
T0 = 300.0

# For mode == 'single'
sigma0 = 95.0      # S/cm
S0     = 28.0e-6   # V/S

# For mode == 'X' and 'T'
# Input and output files
infile  = 'LaTiO3-sigma-S-T.xlsx'
outfile = None

# Indexes of columns for X variable, T, sigma (S/cm), and S (V/K)
X_label     = 0
T_label     = 0
sigma_label = 1
S_label     = 2


#===================================
# figure configuration
#===================================
figsize         = (12, 8)
fontsize        = 16
legend_fontsize = 12
markersize      = 8.0

#===================================
# Treat arguments
#===================================
def pint(str, defval = None):
    try:
        return int(str)
    except:
        return defval

def pfloat(str, defval = None):
    try:
        return float(str)
    except:
        return defval

def getarg(idx, defval = None):
    try:
        return sys.argv[idx]
    except:
        return defval

def getfloatarg(idx, defval = None):
    try:
        return pfloat(sys.argv[idx], defval)
    except:
        return defval

def getintarg(idx, defval = None):
    try:
        return pint(sys.argv[idx], defval)
    except:
        return defval

def modify_path(path, addpath):
    basename    = os.path.basename(path)
    dirname     = os.path.dirname(path)
    header, ext = os.path.splitext(path)
    filebody    = os.path.basename(header)
    return os.path.join(dirname, filebody + addpath)

def usage(app = None):
    argv = sys.argv
    print("")
    print("Usage: python {} single sigma(S/cm) S(V/K) T(K)".format(argv[0]))
    print("   ex: python {} single {} {} {}".format(argv[0], sigma0, S0, T0))
    print("Usage: python {} X infile T(K) X_label sigma_label S_label".format(argv[0]))
    print("   ex: python {} X {} {} {} {} {}".format(argv[0], infile, T0, X_label, sigma_label, S_label))
    print("Usage: python {} T infile iT isigma iS".format(argv[0]))
    print("   ex: python {} T {} {} {} {}".format(argv[0], infile, T_label, sigma_label, S_label))

def updatevars():
    global mode
    global T0
    global sigma0, S0
    global infile, outfile
    global X_label, T_label, sigma_label, S_label

    argv = sys.argv
    mode = getarg( 1, mode)
    if mode == 'single':
        sigma0 = getfloatarg( 2, sigma0)
        S0     = getfloatarg( 3, S0)
        T0     = getfloatarg( 4, T0)
    elif mode == 'X':
        infile       = getarg     ( 2, infile)
        T0           = getfloatarg( 3, T0)
        X_label      = getarg( 4, X_label)
        sigma_label  = getarg( 5, sigma_label)
        S_label      = getarg( 6, S_label)

        outfile = modify_path(infile, '-out.xlsx')
    elif mode == 'T':
        infile       = getarg( 2, infile)
        T_label      = getarg( 3, T_label)
        sigma_label  = getarg( 4, sigma_label)
        S_label      = getarg( 5, S_label)

        outfile = modify_path(infile, '-out.xlsx')
    else:
        app.terminate("Error: Invalid mode [{}]".format(mode), usage = usage, pause = True)


#===================================
# Other functions
#===================================
def read_file(fname):
    print("")

    datafile = tkVariousData(infile)
    labels, datalist = datafile.Read_minimum_matrix(close_fp = True, usage = usage)

    return datafile, labels, datalist


# sigma (S/m), S (V/K)
# muw (m^2/(Vs)), nw (m^-3)
def weighted_mobility(sigma, S, T):
    if S < 0.0:
        sign = -1
        carriertype = 'e'
    else:
        sign = 1
        carriertype = 'h'

    absS  = abs(S)
    kBe   = kB / e
    SkBe = absS / kBe

    m2kBT = 2.0 * me * kB * T
    A1b = 8.0 * pi * e * pow(m2kBT, 1.5)
    A1 = 3.0 * h**3 * sigma / A1b
    A2 = 3.0 / (pi * pi) / kBe * absS
    expp1m2 = exp(SkBe - 2.0)
    expm5m1 = exp(-5.0 * (SkBe - 1.0))
    expp5m1 = exp( 5.0 * (SkBe - 1.0))
    
    muw = A1 * (expp1m2 / (1.0 + expm5m1) + A2 / (1.0 + expp5m1))
    nw  = sigma / e / muw

    return muw, nw, sign, carriertype
    
def exec_single():
    print("")
    print("sigma = {} S/cm".format(sigma0))
    print("S     = {} V/K".format(S0))
    print("T     = {} K".format(T0))

    mu_w, n_w, sign, carriertype = weighted_mobility(sigma0 * 100.0, S0, T0)
    print("")
    print("Major carrier type:", carriertype)
    print("mu_w = {:12.6g} m^2/(Vs) = {:12.6g} cm^2/(Vs)".format(mu_w, mu_w * 1.0e4))
    print("n_w  = {:12.6g} m^-3     = {:12.6g} cm^-3".format(n_w, n_w * 1.0e-6))
    
    print("")
    app.terminate(pause = True)

def validate_data(list, desc):
    if list is None:
        app.terminate(f"\nError in validate_data: Null data for {desc}\n", pause = True)

def exec_X():
    print("")
    print("input file :", infile)
    print("output file:", outfile)
    print("T     = {} K".format(T0))
    print("X_label    :", T_label)
    print("sigma_label:", sigma_label)
    print("S_label    :", S_label)

    print("")
    print("Read data from [{}]".format(infile))
    datafile, labels, datalist = read_file(infile)
    Xlabel, Xlist         = datafile.FindDataArray(X_label, flag = 'i')
    sigmalabel, sigmalist = datafile.FindDataArray(sigma_label, flag = 'i')
    Slabel, Slist         = datafile.FindDataArray(S_label, flag = 'i')
    validate_data(Xlist, 'X data')
    validate_data(sigmalist, 'sigma data')
    validate_data(Slist, 'S data')

    muwlist = []
    nwlist  = []
    ctlist  = []
    print("{:12}\t{:12}\t{:12}\t{:12}\t{:12}\t{}".format(Xlabel, sigmalabel, Slabel, "mu_w(cm^2/(Vs))", "n_w(cm^-3)", 'type'))
    for ic in range(len(datalist)):
        mu_w, n_w, sign, carriertype = weighted_mobility(sigmalist[ic] * 100.0, Slist[ic], T0)
        mu_w *= 1.0e4
        n_w  *= 1.0e-6
        muwlist.append(mu_w)
        nwlist.append(n_w)
        ctlist.append(carriertype)
        print("{:12.6g}\t{:12.6g}\t{:12.6g}\t{:12.6g}\t{:12.6g}\t{}".format(Xlist[ic], sigmalist[ic], Slist[ic], mu_w, n_w, carriertype))

    print("")
    print("Save results to [{}]".format(outfile))
    tkVariousData().to_excel(outfile, 
            [Xlabel, sigmalabel, Slabel, "mu_w(cm^2/(Vs))", "n_w(cm^-3)", 'type'],
            [Xlist, sigmalist, Slist, muwlist, nwlist, ctlist])


#=============================
# Plot graphs
#=============================
    fig = plt.figure(figsize = figsize)

    ax1a = fig.add_subplot(1, 2, 1)
    ax1b = ax1a.twinx()
    ax2a = fig.add_subplot(1, 2, 2)
    ax2b = ax2a.twinx()

    xSp = []
    xSm = []
    Sp  = []
    Sm  = []
    for i in range(len(Xlist)):
        if Slist[i] >= 0.0:
            xSp.append(Xlist[i])
            Sp.append(Slist[i])
        else:
            xSm.append(Xlist[i])
            Sm.append(Slist[i])
    xnp = []
    xnm = []
    np  = []
    nm  = []
    for i in range(len(Xlist)):
        if ctlist[i] == 'h':
            xnp.append(Xlist[i])
            np.append(nwlist[i])
        else:
            xnm.append(Xlist[i])
            nm.append(nwlist[i])
    
    ins1 = ax1a.plot(Xlist, sigmalist, label = sigmalabel,
            linestyle = '', marker = '^', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ins2 = ax1b.plot(xSp, Sp,     label = "{} (h)".format(Slabel),
            linestyle = '', marker = 'o', markerfacecolor = 'blue',  markeredgecolor = 'blue',  markersize = markersize)
    ins3 = ax1b.plot(xSm, Sm,     label = "{} (e)".format(Slabel),
            linestyle = '', marker = 'o', markerfacecolor = 'red',   markeredgecolor = 'red',   markersize = markersize)
    xrange = ax1b.get_xlim()
    ax1b.plot(xrange, [0.0, 0.0], linestyle = 'dashed', color = 'red', linewidth = 0.5)
    ax1a.set_yscale('log')
    ax1a.set_xlabel(Xlabel,     fontsize = fontsize)
    ax1a.set_ylabel(sigmalabel, fontsize = fontsize)
    ax1b.set_ylabel(Slabel,     fontsize = fontsize)
    ax1a.set_xlim(xrange)
    ax1b.set_xlim(xrange)
    ins = ins1 + ins2 + ins3
    ax1a.legend(ins, [l.get_label() for l in ins], fontsize = legend_fontsize, loc = 'upper center') #loc = 'best')
#    h1a, l1a = ax1a.get_legend_handles_labels()
#    h1b, l1b = ax1b.get_legend_handles_labels()
#    ax1a.legend(h1a + h1b, l1a + h1b, fontsize = legend_fontsize)
    ax1a.set_title("{} (T = {} K)".format(infile, T0))
    ax1a.tick_params(labelsize = fontsize)
    ax1b.tick_params(labelsize = fontsize)

    ins1 = ax2a.plot(Xlist, muwlist, label = '$\mu_w$',
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ins2 = ax2b.plot(xnp, np,  label = '$n_w$ (h)',
            linestyle = '', marker = '^', markerfacecolor = 'blue',  markeredgecolor = 'blue',  markersize = markersize)
    ins3 = ax2b.plot(xnm, nm,  label = '$n_w$ (e)',
            linestyle = '', marker = '^', markerfacecolor = 'red',   markeredgecolor = 'red',   markersize = markersize)
    ax2b.plot(xrange, [0.0, 0.0], color = 'red', linestyle = 'dashed', linewidth = 0.5)
    ax2b.set_yscale('log')
    ax2a.set_xlabel(Xlabel,    fontsize = fontsize)
    ax2a.set_ylabel('$\mu_w$ (cm$^2$/(Vs))', fontsize = fontsize)
    ax2b.set_ylabel('$n_w$ (cm$^{-3}$)',   fontsize = fontsize)
    ax2a.set_xlim(xrange)
    ax2b.set_xlim(xrange)
    ins = ins1 + ins2 + ins3
    ax2b.legend(ins, [l.get_label() for l in ins], fontsize = legend_fontsize, loc = 'upper center') #loc = 'best')
#    h2a, l2a = ax2a.get_legend_handles_labels()    ins = ins1 + ins2 
#    h2b, l2b = ax2b.get_legend_handles_labels()
#    ax2a.legend(h2a + h2b, l2a + h2b, fontsize = legend_fontsize)
    ax2a.tick_params(labelsize = fontsize)
    ax2b.tick_params(labelsize = fontsize)

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    plt.pause(0.1)

    app.terminate("", usage = usage, pause = True)

def exec_T():
    global X_label, T_label, sigma_label, S_label

    print("")
    print("input file :", infile)
    print("output file:", outfile)
    print("X_label    :", T_label)
    print("sigma_label:", sigma_label)
    print("S_label    :", S_label)

    print("")
    print("Read data from [{}]".format(infile))
    datafile, labels, datalist = read_file(infile)
    Tlabel, Tlist         = datafile.FindDataArray(T_label, flag = 'i')
    sigmalabel, sigmalist = datafile.FindDataArray(sigma_label, flag = 'i')
    Slabel, Slist         = datafile.FindDataArray(S_label, flag = 'i')
    validate_data(Tlist, 'T data')
    validate_data(sigmalist, 'sigma data')
    validate_data(Slist, 'S data')

    muwlist = []
    nwlist  = []
    ctlist  = []
    print("{:12}\t{:12}\t{:12}\t{:12}\t{:12}\t{}".format(Tlabel, sigmalabel, Slabel, "mu_w(cm^2/(Vs))", "n_w(cm^-3)", 'type'))
    for ic in range(len(Tlist)):
        mu_w, n_w, sign, carriertype = weighted_mobility(sigmalist[ic] * 100.0, Slist[ic], Tlist[ic])
        mu_w *= 1.0e4
        n_w  *= 1.0e-6
        muwlist.append(mu_w)
        nwlist.append(n_w)
        ctlist.append(carriertype)
        print("{:12.6g}\t{:12.6g}\t{:12.6g}\t{:12.6g}\t{:12.6g}\t{}".format(Tlist[ic], sigmalist[ic], Slist[ic], mu_w, n_w, carriertype))

    print("")
    print("Save results to [{}]".format(outfile))
    tkVariousData().to_excel(outfile, 
            [Tlabel, sigmalabel, Slabel, "mu_w(cm^2/(Vs))", "n_w(cm^-3)", 'type'],
            [Tlist, sigmalist, Slist, muwlist, nwlist, ctlist])


#=============================
# Plot graphs
#=============================
    fig = plt.figure(figsize = figsize)

    ax1a = fig.add_subplot(1, 2, 1)
    ax1b = ax1a.twinx()
    ax2a = fig.add_subplot(1, 2, 2)
    ax2b = ax2a.twinx()

    xSp = []
    xSm = []
    Sp  = []
    Sm  = []
    for i in range(len(Tlist)):
        if Slist[i] >= 0.0:
            xSp.append(Tlist[i])
            Sp.append(Slist[i])
        else:
            xSm.append(Tlist[i])
            Sm.append(Slist[i])
    xnp = []
    xnm = []
    np  = []
    nm  = []
    for i in range(len(Tlist)):
        if ctlist[i] == 'h':
            xnp.append(Tlist[i])
            np.append(nwlist[i])
        else:
            xnm.append(Tlist[i])
            nm.append(nwlist[i])
    
    ins1 = ax1a.plot(Tlist, sigmalist, label = sigmalabel,
            linestyle = '', marker = '^', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ins2 = ax1b.plot(xSp, Sp,     label = "{} (h)".format(Slabel),
            linestyle = '', marker = 'o', markerfacecolor = 'blue',  markeredgecolor = 'blue',  markersize = markersize)
    ins3 = ax1b.plot(xSm, Sm,     label = "{} (e)".format(Slabel),
            linestyle = '', marker = 'o', markerfacecolor = 'red',   markeredgecolor = 'red',   markersize = markersize)
    xrange = ax1b.get_xlim()
    ax1b.plot(xrange, [0.0, 0.0], linestyle = 'dashed', color = 'red', linewidth = 0.5)
    ax1a.set_yscale('log')
    ax1a.set_xlabel(Tlabel,     fontsize = fontsize)
    ax1a.set_ylabel(sigmalabel, fontsize = fontsize)
    ax1b.set_ylabel(Slabel,     fontsize = fontsize)
    ax1a.set_xlim(xrange)
    ax1b.set_xlim(xrange)
    ins = ins1 + ins2 + ins3
    ax1a.legend(ins, [l.get_label() for l in ins], fontsize = legend_fontsize, loc = 'upper center') #loc = 'best')
#    h1a, l1a = ax1a.get_legend_handles_labels()
#    h1b, l1b = ax1b.get_legend_handles_labels()
#    ax1a.legend(h1a + h1b, l1a + h1b, fontsize = legend_fontsize)
    ax1a.set_title("{} (T = {} K)".format(infile, T0))
    ax1a.tick_params(labelsize = fontsize)
    ax1b.tick_params(labelsize = fontsize)

    ins1 = ax2a.plot(Tlist, muwlist, label = '$\mu_w$',
            linestyle = '', marker = 'o', markerfacecolor = 'black', markeredgecolor = 'black', markersize = markersize)
    ins2 = ax2b.plot(xnp, np,  label = '$n_w$ (h)',
            linestyle = '', marker = '^', markerfacecolor = 'blue',  markeredgecolor = 'blue',  markersize = markersize)
    ins3 = ax2b.plot(xnm, nm,  label = '$n_w$ (e)',
            linestyle = '', marker = '^', markerfacecolor = 'red',   markeredgecolor = 'red',   markersize = markersize)
    ax2b.plot(xrange, [0.0, 0.0], color = 'red', linestyle = 'dashed', linewidth = 0.5)
    ax2b.set_yscale('log')
    ax2a.set_xlabel(Tlabel,    fontsize = fontsize)
    ax2a.set_ylabel('$\mu_w$ (cm$^2$/(Vs))', fontsize = fontsize)
    ax2b.set_ylabel('$n_w$ (cm$^{-3}$)',   fontsize = fontsize)
    ax2a.set_xlim(xrange)
    ax2b.set_xlim(xrange)
    ins = ins1 + ins2 + ins3
    ax2a.legend(ins, [l.get_label() for l in ins], fontsize = legend_fontsize, loc = 'upper center') #loc = 'best')
#    h2a, l2a = ax2a.get_legend_handles_labels()
#    h2b, l2b = ax2b.get_legend_handles_labels()
#    ax2a.legend(h2a + h2b, l2a + h2b, fontsize = legend_fontsize)
    ax2a.tick_params(labelsize = fontsize)
    ax2b.tick_params(labelsize = fontsize)

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    plt.pause(0.1)

    app.terminate("", usage = usage, pause = True)


def main():
    global app

    app = tkApplication()
#    logfile = app.replace_path(infile)
#    print(f"Open logfile [{logfile}]")
#    app.redirect(targets = ["stdout", logfile], mode = 'w')

    print("")
    print("=============================================================")
    print("Calculate weighted mobility and carrier concentration ")
    print("from conductivity (S/cm) and Seebeck coefficient (V/K)")
    print("=============================================================")
    print("")

    updatevars()

    print("mode:", mode)
    
    if mode == 'single':
        exec_single()
    elif mode == 'X':
        exec_X()
    elif mode == 'T':
        exec_T()
    else:
        app.terminate("Error: Invalid mode [{}]".format(mode), usage = usage, pause = True)


if (__name__ == '__main__'):
    main()
